# Copyright (c) 2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Text formatting for GWpy plots."""

from __future__ import annotations

from typing import TYPE_CHECKING

from astropy.units import UnitBase
from matplotlib import rcParams

from . import tex

if TYPE_CHECKING:
    from matplotlib.axis import Axis

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def to_string(
    input_: str | float | UnitBase,
) -> str:
    """Format an input for representation as text.

    This method is just a convenience that handles default LaTeX formatting.
    """
    usetex = rcParams["text.usetex"]
    if isinstance(input_, UnitBase):
        return input_.to_string("latex_inline")
    if isinstance(input_, float | int) and usetex:
        return tex.float_to_latex(input_)
    if usetex:
        return tex.label_to_latex(input_)
    return str(input_)


def default_unit_label(
    axis: Axis,
    unit: UnitBase,
    format: str = "latex_inline_dimensional",  # noqa: A002
) -> str:
    """Set default label for an axis from a `~astropy.units.Unit`.

    If the axis already has a label, this function does nothing except
    return the axis label.

    Parameters
    ----------
    axis : `~matplotlib.axis.Axis`
        The axis to manipulate.

    unit : `~astropy.units.Unit`
        The unit to use for the label.

    format : `str`, optional
        The format to use when converting the unit to a label.

    Returns
    -------
    text : `str`
        The text for the new label, if set.
    """
    if axis.isDefault_label:
        axis.set_label_text(unit.to_string(format=format))
        axis.isDefault_label = True
    return axis.get_label_text()



# Copyright (c) 2015-2020 Joseph Areeda
#               2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""The timeseries CLI product."""

from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING

from ..plot import Plot
from ..plot.tex import label_to_latex
from .cliproduct import TimeDomainProduct

if TYPE_CHECKING:
    from argparse import Namespace
    from logging import Logger
    from typing import ClassVar

    from ..plot import Axes

__author__ = "Joseph Areeda <joseph.areeda@ligo.org>"

logger = logging.getLogger(__name__)


class TimeSeriesProduct(TimeDomainProduct):
    """Plot one or more time series."""

    action: ClassVar[str] = "timeseries"

    def __init__(
        self,
        args: Namespace,
        logger: Logger = logger,
    ) -> None:
        """Create a new `TimeSeriesProduct`."""
        super().__init__(args, logger=logger)

    def get_ylabel(self) -> str | None:
        """Text for y-axis label,  check if channel defines it."""
        units = self.units
        if len(units) == 1 and str(units[0]) == "":  # dimensionless
            return ""
        if len(units) == 1 and self.usetex:
            return units[0].to_string("latex")
        if len(units) == 1:
            return units[0].to_string()
        if len(units) > 1:
            return "Multiple units"
        return super().get_ylabel()

    def get_suptitle(self) -> str:
        """Start of default super title, first channel is appended to it."""
        return f"Time series: {self.chan_list[0]}"

    def get_title(self) -> str:
        """Return the title of this `TimeSeries` product."""
        suffix = super().get_title()
        # limit significant digits for minute trends
        rates = {ts.sample_rate.round(3) for ts in self.timeseries}
        fss = f"({'), ('.join(map(str, rates))})"
        return ", ".join([
            f"Fs: {fss}",
            f"duration: {self.duration}",
            suffix,
        ])

    def make_plot(self) -> Plot:
        """Generate the plot from time series and arguments."""
        plot = Plot(figsize=self.figsize, dpi=self.dpi)
        ax = plot.add_subplot(xscale="auto-gps")

        # handle user specified plot labels
        if self.args.legend:
            nlegargs = len(self.args.legend[0])
        else:
            nlegargs = 0
        if nlegargs > 0 and nlegargs != self.n_datasets:
            warnings.warn(
                "The number of legends specified must match the number of "
                "time series (channels * start times). "
                f"There are {len(self.timeseries)} series "
                f"and {len(self.args.legend)} legends",
                stacklevel=2,
            )
            nlegargs = 0    # don't use them

        # get colours
        colors = self._color_by_ifo()

        for i in range(self.n_datasets):
            series = self.timeseries[i]
            if nlegargs:
                label = self.args.legend[0][i]
            elif series.channel:  # GWOSC data doesn't have a 'channel'
                label = series.channel.name
            else:
                label = series.name
            if self.usetex:
                label = label_to_latex(label)
            ax.plot(series, label=label, color=colors[i])

        return plot

    def _scale_axes_from_data(self, ax: Axes) -> None:
        """Restrict data limits for Y-axis based on what you can see."""
        # get tight limits for X-axis
        if self.args.xmin is None:
            self.args.xmin = min(ts.xspan[0] for ts in self.timeseries)
        if self.args.xmax is None:
            self.args.xmax = max(ts.xspan[1] for ts in self.timeseries)

        # autoscale view for Y-axis
        cropped = [ts.crop(self.args.xmin, self.args.xmax) for
                   ts in self.timeseries]
        ymin = min(ts.value.min() for ts in cropped)
        ymax = max(ts.value.max() for ts in cropped)
        ax.yaxis.set_data_interval(ymin, ymax, ignore=True)
        ax.autoscale_view(scalex=False)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.timeseries

Plotting an averaged ASD with percentiles.
##########################################

As we have seen in :doc:`hoff`, the Amplitude Spectral Density (ASD) is a key
indicator of frequency-domain sensitivity for a gravitational-wave detector.

However, the ASD doesn't allow you to see how much the sensitivity varies
over time.
One tool for that is the :doc:`spectrogram </spectrogram/index>`, while another
is simply to show percentiles of a number of ASD measurements.

In this example we calculate the median ASD over 2048-seconds surrounding
the GW178017 event, and also the 5th and 95th percentiles of the ASD, and
plot them on a single figure.
"""

# %%
# Data access
# -----------
# First, as always, we get the data using :meth:`TimeSeries.get`:

from gwpy.timeseries import TimeSeries
hoft = TimeSeries.get("H1", 1187007040, 1187009088)

# %%
# Calculate spectrogram
# ---------------------
# Next we calculate a :class:`~gwpy.spectrogram.Spectrogram` by calculating
# a number of ASDs, using the :meth:`~gwpy.timeseries.TimeSeries.spectrogram2`
# method:

sg = hoft.spectrogram2(fftlength=4, overlap=2, window="hann") ** (1 / 2.)

# %%
# From this we can trivially extract the median, 5th and 95th percentiles:

median = sg.percentile(50)
low = sg.percentile(5)
high = sg.percentile(95)

# %%
# Visualisation
# -------------
# Finally, we can make plot, using :meth:`~gwpy.plot.Axes.plot_mmm` to
# display the 5th and 95th percentiles as a shaded region around the median:

from gwpy.plot import Plot
plot = Plot()
ax = plot.add_subplot(
    xscale="log",
    xlim=(10, 1500),
    xlabel="Frequency [Hz]",
    yscale="log",
    ylim=(3e-24, 2e-20),
    ylabel=r"Strain noise [1/$\sqrt{\mathrm{Hz}}$]",
)
ax.plot_mmm(median, low, high, color="gwpy:ligo-hanford")
ax.set_title("LIGO-Hanford strain noise variation around GW170817")
plot.show()

# %%
# Now we can see that the ASD varies by factors of a few across most of the
# frequency band, with notable exceptions, e.g. around the 60-Hz power line
# harmonics (60 Hz, 120 Hz, 180 Hz, ...) where the noise is very stable.



# Copyright (c) 2024-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""I/O utilities for GWF files.

Most of the actual GWF file I/O is delegated to one of the 'backend'
implementations that rely on a third-party library.
"""

# support for GWF I/O backends
from .backend import *

# basic functionality that should be underpinned by a backend implementation
from .core import *

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# -- Module-level variable docstrings
# Python doesn't support docstrings for variables, and the autodoc
# standard for module-level variables doesn't carry across imports,
# so we use this workaround.

#: Preferentially ordered list of supported GWF backends.
#:
#: When GWF functions are called without an explicit backend argument,
#: these backends will be tried in order until one is found that
#: provides the requested functionality.
BACKENDS = BACKENDS  # noqa: F405, PLW0127



# Copyright (c) 2019-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Test suite for all examples."""

import os
import re
import warnings
from contextlib import contextmanager
from functools import wraps
from pathlib import Path

import pytest
from matplotlib import use

from gwpy.io.nds2 import NDSWarning
from gwpy.testing.errors import pytest_skip_flaky_network

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

use("agg")  # force non-interactive backend

# find all examples
EXAMPLE_BASE = Path(__file__).parent
EXAMPLE_DIRS = [exdir for exdir in EXAMPLE_BASE.iterdir() if exdir.is_dir()]
EXAMPLES = sorted([
    pytest.param(expy, id=str(expy.relative_to(EXAMPLE_BASE))) for
    exdir in EXAMPLE_DIRS for expy in exdir.glob("*.py")],
)


@contextmanager
def cwd(path):
    """Context manager to change to a given directory and back again."""
    oldpwd = Path.cwd()
    os.chdir(str(path))
    try:
        yield
    finally:
        os.chdir(str(oldpwd))


# -- fixtures ---------------

@pytest.fixture(autouse=True)
def close_figures():
    """Fixture to close all matplotlib figures after each test."""
    from matplotlib import pyplot
    try:
        yield
    finally:
        # close all open figures regardless of test status
        pyplot.close("all")


# acceptable authentication failures
NDS2_AUTH_FAILURES = [
    "failed to establish a connection",
    "error authenticating against",
]
NDS2_SKIP = re.compile(
    "({})".format("|".join(NDS2_AUTH_FAILURES)),
    re.IGNORECASE,
)

# stderr message for NDS2 data access failure
NDS2_READ_ERROR_STDERR = re.compile(r"read_server_response: Wrong length read")


def skip_nds_authentication_error(func):
    """Ignore NDS2 authentication errors."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            with warnings.catch_warnings():  # force NDS warnings as errors
                for msg in NDS2_AUTH_FAILURES:
                    warnings.filterwarnings(
                        "error",
                        message=msg,
                        category=NDSWarning,
                    )
                return func(*args, **kwargs)  # run the test
        except NDSWarning as exc:  # pragma: no-cover
            # if we can't authenticate, dont worry
            if NDS2_SKIP.match(str(exc)):
                pytest.skip(str(exc))
            raise

    return wrapper


def skip_missing_optional_dependency(func):
    """Ignore missing optional dependencies (mainly in `TimeSeries.get`)."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)  # run the test
        except (
            ImportError,  # direct import error
            RuntimeError,  # chained error
        ) as exc:
            msg = str(exc)
            cause = exc.__cause__
            while cause:  # walk up the error chain
                msg += f" caused by '{cause}'"
                cause = cause.__cause__
            # needs an optional dependency
            if not isinstance(exc, ImportError):
                raise
            pytest.skip(msg)

    return wrapper


# -- test -------------------

@pytest.mark.parametrize("script", EXAMPLES)
@pytest.mark.filterwarnings(
    "ignore:Matplotlib is currently using agg",
    "ignore:.*Agg is non-interactive",
)
@pytest_skip_flaky_network  # if there are network errors, we don't care
@skip_nds_authentication_error
@skip_missing_optional_dependency
def test_example(capfd, script):
    """Test an example script runs without error."""
    # read example code from file
    code = compile(
        script.read_text(),
        str(script),
        "exec",
    )

    # in the directory of the script, run it
    try:
        with cwd(script.parent):
            exec(code, globals())
    except* RuntimeError as exc:  # NDS2 failed
        stderr = capfd.readouterr().err
        if NDS2_READ_ERROR_STDERR.match(stderr):
            pytest.skip(f"{exc} caused by '{NDS2_READ_ERROR_STDERR.pattern}'")
        raise



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.log`."""

import logging
import os
import sys
from importlib import reload
from unittest import mock

import pytest

from gwpy import log as gwpy_log


@mock.patch.dict("os.environ", clear=True)
@pytest.mark.parametrize(("env", "expected"), [
    pytest.param(None, logging.NOTSET, id="default"),
    pytest.param("DEBUG", logging.DEBUG, id="DEBUG"),
    pytest.param("debug", logging.DEBUG, id="debug"),
    pytest.param(10, logging.DEBUG, id="debug-int"),
    pytest.param("INFO", logging.INFO, id="INFO"),
    pytest.param(20, logging.INFO, id="info-int"),
    pytest.param("WARNING", logging.WARNING, id="WARNING"),
    pytest.param("ERROR", logging.ERROR, id="ERROR"),
    pytest.param("CRITICAL", logging.CRITICAL, id="CRITICAL"),
])
def test_get_default_level(env, expected):
    """Test :func:`gwpy.log.get_default_level`."""
    if env is not None:
        os.environ["GWPY_LOG_LEVEL"] = str(env)
    # Check that we get the expected level string
    assert gwpy_log.get_default_level() is expected


def test_init_logger():
    """Test :func:`gwpy.log.get_logger`."""
    # Basic test of getting a logger
    logger = gwpy_log.init_logger("gwpy.log")
    assert logger.name == "gwpy.log"
    # Check that we get the same logger if we call again
    assert gwpy_log.init_logger("gwpy.log") is logger


@mock.patch.dict("os.environ")
@pytest.mark.parametrize("env", [
    pytest.param({}, id="default"),
    pytest.param({
        "GWPY_LOG_LEVEL": "DEBUG",
    }, id="debug"),
    pytest.param({
        "GWPY_LOG_FORMAT": "%(name)s: %(message)s",
    }, id="format"),
    pytest.param({
        "GWPY_LOG_DATEFMT": "%H:%M:%S",
    }, id="datefmt"),
    pytest.param({
        "GWPY_LOG_LEVEL": "DEBUG",
        "GWPY_LOG_FORMAT": "%(name)s: %(message)s",
        "GWPY_LOG_DATEFMT": "%H:%M:%S",
    }, id="all"),
])
def test_init_logger_env(env):
    """Test :func:`gwpy.log.init_logger` with environment variables."""
    # Reset the test logger
    name = __name__
    logging.getLogger(name).handlers.clear()
    logging.getLogger(name).propagate = False

    # Update the env
    os.environ.update(env)

    # Reload the log module to reset any cached values
    reload(gwpy_log)

    # Check that the logger is initialised properly
    logger = gwpy_log.init_logger(name)
    assert logger.name == name

    if "GWPY_LOG_LEVEL" in env:
        assert logger.level == logging.DEBUG
    else:
        assert logger.level == logging.NOTSET

    handler = logger.handlers[0]
    assert (formatter := handler.formatter)
    assert formatter._fmt == os.getenv(
        "GWPY_LOG_FORMAT",
        gwpy_log.DEFAULT_LOG_FORMAT,
    )
    assert formatter.datefmt == os.getenv(
        "GWPY_LOG_DATEFMT",
        gwpy_log.DEFAULT_LOG_DATEFMT,
    )


# -- Test package-level log initialisation

@mock.patch.dict("sys.modules")
@mock.patch("gwpy.init_logging")
def test_package_init_logging(init_logging):
    """Test that we don't initialise logging at the package level by default."""
    del sys.modules["gwpy"]
    # By default the gwpy logger should not be configured.
    import gwpy  # noqa: F401
    init_logging.assert_not_called()


def test_package_init_logging_env():
    """Test that we can initialise logging at the package level."""
    del sys.modules["gwpy"]
    os.environ["GWPY_INIT_LOGGING"] = "1"
    # By default the gwpy logger should not be configured.
    import gwpy  # noqa: F401
    logger = logging.getLogger("gwpy")
    assert logger.level == logging.INFO



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Quantity array for indexing a Series."""

from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

import numpy
from astropy.units import Quantity

from .array import COPY_IF_NEEDED

if TYPE_CHECKING:
    from typing import Self

    from numpy.typing import (
        ArrayLike,
        DTypeLike,
    )


class Index(Quantity):
    """1-D `~astropy.units.Quantity` array for indexing a `Series`.

    See Also
    --------
    astropy.units.Quantity
        For parameters supported when creating an `Index` array.
    """

    @classmethod
    def define(
        cls,
        start: float,
        step: float,
        num: int,
        dtype: DTypeLike = None,
    ) -> Self:
        """Define a new `Index` using `numpy.arange`.

        The output is basically::

            start + numpy.arange(num) * step

        Parameters
        ----------
        start : `Number`
            The starting value of the index.

        step : `Number`
            The step size of the index.

        num : `int`
            The size of the index (number of samples).

        dtype : `numpy.dtype`, `None`, optional
            The desired dtype of the index, if not given, defaults
            to the higher-precision dtype from ``start`` and ``step``.

        Returns
        -------
        index : `Index`
            A new `Index` created from the given parameters.
        """
        if dtype is None:
            dtype = max(
                numpy.array(start, subok=True, copy=COPY_IF_NEEDED).dtype,
                numpy.array(step, subok=True, copy=COPY_IF_NEEDED).dtype,
            )
        startq = Quantity(start, dtype=dtype, copy=COPY_IF_NEEDED)
        stepq = Quantity(step, dtype=dtype, copy=COPY_IF_NEEDED).to(startq.unit)
        stopq = startq + stepq * num
        return cls(
            numpy.arange(
                startq.value,
                stopq.value,
                stepq.value,
                dtype=dtype,
            )[:num],
            unit=startq.unit,
            copy=COPY_IF_NEEDED,
        )

    @property
    def regular(self) -> bool:
        """`True` if this index is linearly increasing."""
        try:
            return self.info.meta["regular"]
        except (TypeError, KeyError):
            if self.info.meta is None:
                self.info.meta = {}
            self.info.meta["regular"] = self.is_regular()
            return self.info.meta["regular"]

    def is_regular(self) -> bool:
        """Determine whether this `Index` contains linearly increasing samples.

        This also works for linear decrease.
        """
        if self.size <= 1:
            return False
        return bool(numpy.isclose(numpy.diff(self.value, n=2), 0).all())

    def __getitem__(
        self,
        key: slice | int | bool | ArrayLike,
    ) -> Quantity:
        """Get an item or a slice of this `Index`."""
        item = super().__getitem__(key)
        if item.isscalar:
            return item.view(Quantity)
        return item

    def __setitem__(
        self,
        key: slice | int | bool | ArrayLike,
        value: float | ArrayLike,
    ) -> None:
        """Set an element or slice of this `Index`."""
        super().__setitem__(key, value)
        # We have changed the Index, so reset our understanding of regularity.
        with contextlib.suppress(AttributeError, TypeError):
            self.info.meta.pop("regular", None)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Decorators for GWpy."""

from __future__ import annotations

import warnings
from functools import wraps
from typing import (
    TYPE_CHECKING,
    no_type_check,
)

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import (
        ParamSpec,
        TypeVar,
    )

    P = ParamSpec("P")
    R = TypeVar("R")
    ReturnType = TypeVar("ReturnType")

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "deprecated_property",
    "deprecated_function",
    "return_as",
]

DEPRECATED_FUNCTION_WARNING: str = (
    "{0.__module__}.{0.__name__} has been deprecated, and will be "
    "removed in a future release."
)


class deprecated_property(property):  # noqa: N801
    """Sub-class of `property` that invokes DeprecationWarning on every call.

    .. deprecated:: 4.0.0

        This property is deprecated and will be removed in a future release.
        Use `warnings.deprecated` instead.
    """

    @no_type_check
    def __init__(
        self,
        fget: Callable,
        fset: Callable | None = None,
        fdel: Callable | None = None,
        doc: str | None = None,
    ) -> None:
        """Create a property that will issue a DeprecationWarning."""
        warnings.warn(
            "the deprecated_property decorator is itself deprecated and will be "
            "removed in a future release, please use the warnings.deprecated decorator "
            "from the Python 3.13+ standard library instead",
            category=DeprecationWarning,
            stacklevel=2,
        )

        # get name of property
        pname = fget.__name__

        # build a wrapper that will spawn a DeprecationWarning for all calls
        def _warn(func: Callable[P, R]) -> Callable[P, R]:
            """Wrap a function to issue a DeprecationWarning."""
            @wraps(func)
            def _wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
                """Wrap a function to issue a DeprecationWarning."""
                inst = args[0]
                warnings.warn(
                    f"the {type(inst).__name__}.{pname} property is deprecated "
                    "and will be removed in a future release, please stop "
                    "using it.",
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                return func(*args, **kwargs)

            return _wrapped

        # wrap the property methods
        if fdel:
            fdel = _warn(fdel)
        if fset:
            fset = _warn(fset)
        if not fset and not fdel:  # only wrap once
            fget = _warn(fget)

        super().__init__(fget, fset, fdel, doc)


def deprecated_function(
    func: Callable | None = None,
    message: str = DEPRECATED_FUNCTION_WARNING,
) -> Callable:
    """Add a `DeprecationWarning` to a function.

    .. deprecated:: 4.0.0

        This function is deprecated and will be removed in a future release.
        Use `warnings.deprecated` instead.

    Parameters
    ----------
    func : `callable`
        The function to decorate with a `DeprecationWarning`.

    message : `str`, optional
        The warning message to present.

    Notes
    -----
    The final warning message is formatted using `str.format`,
    e.g ``message.format(func)``, so you can use attribute references
    to the function itself.
    See the default message as an example.
    """
    def _decorator(func: Callable[P, R]) -> Callable[P, R]:
        @wraps(func)
        def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> R:
            warnings.warn(
                message.format(func),
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapped_func

    if func:
        return _decorator(func)
    return _decorator


def return_as(
    returntype: Callable[[R], ReturnType],
) -> Callable[[Callable[P, R]], Callable[P, ReturnType]]:
    """Decorate a function to cast the return as the given type.

    Parameters
    ----------
    returntype : `type`
        the desired return type of the decorated function
    """
    def decorator(func: Callable[P, R]) -> Callable[P, ReturnType]:
        """Decorate a function to cast the return value as the given type."""
        @wraps(func)
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> ReturnType:
            """Wrap a function to cast the return value as the given type."""
            result = func(*args, **kwargs)
            try:
                return returntype(result)
            except (TypeError, ValueError) as exc:
                fname = getattr(func, "__name__", str(func))
                rname = getattr(returntype, "__name__", str(returntype))
                exc.args = (
                    f"failed to cast return from {fname} as {rname}: {exc}",
                )
                raise

        return wrapped

    return decorator



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWF I/O utilities."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import TypeVar

    _T = TypeVar("_T")


def _channel_dict_kwarg(
    value: dict[str, _T] | list[_T] | _T | None,
    channels: list[str],
    expected_type: type[_T] | None = None,
    varname: str = "value list",
) -> dict[str, _T | None]:
    """Format the given kwarg value in a dict with one value per channel.

    Parameters
    ----------
    value : any type
        Keyword argument value as given by user.

    channels : `list` of `str`
        List of channels being read.

    expected_type : `type`, `types.UnionType`, optional
        Valid type (or union of types) for value (or items in an iterable value).

    varname : `str`, optional
        The name of the keyword variable being parsed.
        This is only used to assist in error reporting.

    Returns
    -------
    dict : `dict`
        `dict` of values, one value per channel key
    """
    # if the value given is of the right type already use it for everything
    if (
        expected_type is not None
        and isinstance(value, expected_type)
        # check whether the expected type is a list or tuple, and whether
        # the input value is a list|tuple of lists|tuples
        and not (
            issubclass(expected_type, list | tuple)
            and len(value)
            and isinstance(value[0], expected_type)
        )
    ):
        # copy value for all channels
        return dict.fromkeys(channels, value)

    # zip list of values with list of channels
    if isinstance(value, list | tuple):
        try:
            return dict(zip(channels, value, strict=True))
        except ValueError as exc:
            exc.args = (
                str(exc).replace(
                    "argument 2", varname,
                ).replace("argument 1", "channels list"),
            )
            raise

    # return subset of value dict for channels
    if isinstance(value, dict):
        return {c: value[c] for c in set(channels).intersection(value)}

    # repeat value for all channels
    return dict.fromkeys(channels, value)



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Extension of `~matplotlib.axes.Axes` for gwpy."""

from __future__ import annotations

import contextlib
from functools import wraps
from math import log
from numbers import Number
from typing import (
    TYPE_CHECKING,
    cast,
)

import numpy
from astropy.time import Time
from matplotlib import (
    _docstring,
    rcParams,
)
from matplotlib.artist import allow_rasterization
from matplotlib.axes import Axes as _Axes
from matplotlib.axes._base import _process_plot_var_args
from matplotlib.collections import PolyCollection
from matplotlib.lines import Line2D
from matplotlib.projections import register_projection

from ..time import to_gps
from .colorbar import colorbar
from .colors import format_norm
from .gps import GPS_SCALES
from .legend import HandlerLine2D

if TYPE_CHECKING:
    from collections.abc import (
        Iterator,
        Sequence,
    )
    from typing import Literal

    import PIL.Image
    from matplotlib.artists import Artist
    from matplotlib.backend_bases import RendererBase
    from matplotlib.collections import (
        Collection,
        PathCollection,
        QuadMesh,
    )
    from matplotlib.colorbar import Colorbar
    from matplotlib.colors import (
        Colormap,
        Normalize,
    )
    from matplotlib.container import BarContainer
    from matplotlib.figure import Figure
    from matplotlib.image import AxesImage
    from matplotlib.legend import Legend
    from matplotlib.patches import Polygon
    from matplotlib.transforms import Bbox
    from matplotlib.typing import ColorType
    from numpy.typing import ArrayLike

    from gwpy.plot.gps import GPSTransform
    from gwpy.time import SupportsToGps
    from gwpy.types import (
        Array2D,
        Series,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def _sortby(
    sortby: ArrayLike,
    *arrays: ArrayLike,
) -> Iterator[numpy.ndarray]:
    """Sort a set of arrays by the first one (including the first one)."""
    # try and sort the colour array by value
    sortidx = numpy.asanyarray(sortby, dtype=float).argsort()

    def _sort(arr: ArrayLike) -> numpy.ndarray:
        if arr is None or isinstance(arr, Number):
            return arr
        return numpy.asarray(arr)[sortidx]

    # apply the sorting to each data array, and scatter
    for arr in (sortby, *arrays):
        yield _sort(arr)


def _poly_ll(x: float, y: float, w: float, h: float) -> tuple[
    tuple[float, float],
    tuple[float, float],
    tuple[float, float],
    tuple[float, float],
]:
    """Return polygon vertices for a rectangle anchored at the lower-left point."""
    return ((x, y), (x, y+h), (x+w, y+h), (x+w, y))


def _poly_lr(x: float, y: float, w: float, h: float) -> tuple[
    tuple[float, float],
    tuple[float, float],
    tuple[float, float],
    tuple[float, float],
]:
    """Return polygon vertices for a rectangle anchored at the lower-right point."""
    return ((x-w, y), (x-w, y+h), (x, y+h), (x, y))


def _poly_ul(x: float, y: float, w: float, h: float) -> tuple[
    tuple[float, float],
    tuple[float, float],
    tuple[float, float],
    tuple[float, float],
]:
    """Return polygon vertices for a rectangle anchored at the upper-left point."""
    return ((x, y-h), (x, y), (x+w, y), (x+w, y-h))


def _poly_ur(x: float, y: float, w: float, h: float) -> tuple[
    tuple[float, float],
    tuple[float, float],
    tuple[float, float],
    tuple[float, float],
]:
    """Return polygon vertices for a rectangle anchored at the upper-right point."""
    return ((x-w, y-h), (x-w, y), (x, y), (x, y-h))


def _poly_center(x: float, y: float, w: float, h: float) -> tuple[
    tuple[float, float],
    tuple[float, float],
    tuple[float, float],
    tuple[float, float],
]:
    """Return polygon vertices for a rectangle anchored at the centre point."""
    return (
        (x-w/2., y-h/2.),
        (x-w/2., y+h/2.),
        (x+w/2., y+h/2.),
        (x+w/2., y-h/2.),
    )


# -- new Axes ------------------------

class Axes(_Axes):
    """GWpy customised `~matplotlib.axes.Axes`."""

    def __init__(
        self,
        fig: Figure,
        *args: tuple[float, float, float, float] | Bbox | int,
        **kwargs,
    ) -> None:
        """Initialise a new `Axes`."""
        super().__init__(fig, *args, **kwargs)

        # handle Series in `ax.plot()`
        self._get_lines = PlotArgsProcessor()

        # reset data formatters (for interactive plots) to support
        # GPS time display
        self.fmt_xdata = self._fmt_xdata
        self.fmt_ydata = self._fmt_ydata

    @wraps(_Axes.draw)
    @allow_rasterization
    def draw(self, renderer: RendererBase) -> None:
        """Draw these `Axes` using the given renderer."""
        # Set the default GPS labels for each axis
        labels = {}
        for ax in (self.xaxis, self.yaxis):
            if ax.get_scale() in GPS_SCALES and ax.isDefault_label:
                labels[ax] = ax.get_label_text()
                trans = cast("GPSTransform", ax.get_transform())
                epoch = float(trans.get_epoch())
                unit = trans.get_unit_name()
                iso = Time(epoch, format="gps", scale="utc").iso
                utc = iso.rstrip("0").rstrip(".")
                ax.set_label_text(f"Time [{unit}] from {utc} UTC ({epoch!r})")

        try:
            super().draw(renderer)
        finally:
            for ax in labels:  # reset labels
                ax.isDefault_label = True

    # -- auto-gps helpers ------------

    def _fmt_xdata(self, x: float) -> str:
        """Format a value for display on the X-Axis."""
        if self.get_xscale() in GPS_SCALES:
            return str(to_gps(x))
        return self.xaxis.get_major_formatter().format_data_short(x)

    def _fmt_ydata(self, y: float) -> str:
        """Format a value for display on the Y-Axis."""
        if self.get_yscale() in GPS_SCALES:
            return str(to_gps(y))
        return self.yaxis.get_major_formatter().format_data_short(y)

    @wraps(_Axes.set_xlim)
    def set_xlim(
        self,
        left: SupportsToGps | tuple[SupportsToGps, SupportsToGps] | None = None,
        right: float | SupportsToGps | None = None,
        **kwargs,
    ) -> tuple[float, float]:
        """Set the X-axis view limits."""
        if right is None and numpy.iterable(left):
            left, right = left
        if self.get_xscale() in GPS_SCALES:
            with contextlib.suppress(TypeError):
                left = numpy.longdouble(str(to_gps(left)))
            with contextlib.suppress(TypeError):
                right = numpy.longdouble(str(to_gps(right)))
        return super().set_xlim(left=left, right=right, **kwargs)

    def set_epoch(self, epoch: SupportsToGps) -> None:
        """Set the epoch for the current GPS scale.

        This method will fail if the current X-axis scale isn't one of
        the GPS scales. See :ref:`gwpy-plot-gps` for more details.

        Parameters
        ----------
        epoch : `float`, `str`
            GPS-compatible time or date object, anything parseable by
            :func:`~gwpy.time.to_gps` is fine.
        """
        scale = self.get_xscale()
        return self.set_xscale(scale, epoch=epoch)

    def get_epoch(self) -> float | None:
        """Return the epoch for the current GPS scale/.

        This method will fail if the current X-axis scale isn't one of
        the GPS scales. See :ref:`gwpy-plot-gps` for more details.
        """
        return self.get_xaxis().get_transform().get_epoch()  # type: ignore[attr-defined]

    # -- overloaded plotting methods -

    def scatter(
        self,
        x: float | ArrayLike,
        y: float | ArrayLike,
        s: float | ArrayLike | None = None,
        c: ArrayLike | Sequence[ColorType] | ColorType | None = None,
        **kwargs,
    ) -> PathCollection:
        """Scatter ``y`` vs ``x`` with varying marker size and/or colour."""
        # This method overloads Axes.scatter to enable quick
        # sorting of data by the 'colour' array before scatter
        # plotting.

        if kwargs.pop("sortbycolor", False) and c is not None:
            # try and sort the colour array by value
            try:
                c, x, y, s = _sortby(c, x, y, s)
            except ValueError as exc:
                exc.args = (
                    "Axes.scatter argument 'sortbycolor' can only be used "
                    "with a simple array of floats in the colour array 'c'",
                )
                raise

        return super().scatter(x, y, s=s, c=c, **kwargs)

    scatter.__doc__ = _Axes.scatter.__doc__.replace(  # type: ignore[union-attr]
        "marker :",
        "sortbycolor : `bool`, optional, default: False\n"
        "    Sort scatter points by `c` array value, if given.\n\n"
        "marker :",
    )

    @_docstring.interpd
    def imshow(  # noqa: D417
        self,
        X: ArrayLike | PIL.Image.Image,  # noqa: N803
        cmap: str | Colormap | None = None,
        norm: str | Normalize | None = None,
        **kwargs,
    ) -> AxesImage:
        """Display an image, i.e. data on a 2D regular raster.

        If ``array`` is a :class:`~gwpy.types.Array2D` (e.g. a
        :class:`~gwpy.spectrogram.Spectrogram`), then the defaults are
        _different_ to those in the upstream
        :meth:`~matplotlib.axes.Axes.imshow` method. Namely, the defaults are

        - ``origin='lower'`` (coordinates start in lower-left corner)
        - ``aspect='auto'`` (pixels are not forced to be square)
        - ``interpolation='none'`` (no image interpolation is used)

        In all other usage, the defaults from the upstream matplotlib method
        are unchanged.

        Parameters
        ----------
        X : array-like or PIL image
            The image data.

        %(cmap_doc)s

        %(norm_doc)s

        *args, **kwargs
            All arguments and keywords are passed to the inherited
            :meth:`~matplotlib.axes.Axes.imshow` method.

        See Also
        --------
        matplotlib.axes.Axes.imshow
            for details of the image rendering
        """
        from gwpy.types import Array2D

        # handle log normalisation
        norm, kwargs = format_norm(kwargs | {"norm": norm})

        # handle Array2D as a special case
        if isinstance(X, Array2D):
            return self._imshow_array2d(X, cmap=cmap, norm=norm, **kwargs)

        # otherwise call back to MPL's Axes.imshow
        image = super().imshow(X, cmap=cmap, norm=norm, **kwargs)
        self.autoscale(enable=None, axis="both", tight=None)
        return image

    def _imshow_array2d(
        self,
        array: Array2D,
        origin: Literal["upper", "lower"] | None = "lower",
        interpolation: str | None = "none",
        aspect: Literal["equal", "auto"] | float | None = "auto",
        **kwargs,
    ) -> AxesImage:
        """Render an `~gwpy.types.Array2D` using `Axes.imshow`."""
        # NOTE: If you change the defaults for this method, please update
        #       the docstring for `imshow` above.

        # calculate extent
        extent = tuple(array.xspan) + tuple(array.yspan)
        if self.get_xscale() == "log" and extent[0] == 0.:
            extent = (1e-300, *extent[1:])
        if self.get_yscale() == "log" and extent[2] == 0.:
            extent = (*extent[:2], 1e-300, *extent[3:])
        kwargs.setdefault("extent", extent)

        return self.imshow(
            array.value.T,
            origin=origin,
            aspect=aspect,
            interpolation=interpolation,
            **kwargs,
        )

    def pcolormesh(
        self,
        *args: ArrayLike,
        **kwargs,
    ) -> QuadMesh:
        """Create a pseudocolor plot with a non-regular rectangular grid.

        When using GWpy, this method can be called with a single argument
        that is an :class:`~gwpy.types.Array2D`, for which the ``X`` and ``Y``
        coordinate arrays will be determined from the indexing.

        In all other usage, all ``args`` and ``kwargs`` are passed directly
        to :meth:`~matplotlib.axes.Axes.pcolormesh`.

        Notes
        -----
        Unlike the upstream :meth:`matplotlib.axes.Axes.pcolormesh`,
        this method respects the current grid settings.

        See Also
        --------
        matplotlib.axes.Axes.pcolormesh
        """
        # handle log normalisation
        norm, kwargs = format_norm(kwargs)

        # handle Array2D as a special case
        if len(args) == 1 and hasattr(args[0], "yindex"):
            array = cast("Array2D", args[0])
            return self._pcolormesh_array2d(array, norm=norm, **kwargs)

        return super().pcolormesh(*args, norm=norm, **kwargs)

    def _pcolormesh_array2d(
        self,
        array: Array2D,
        **kwargs,
    ) -> QuadMesh:
        """Render an `~gwpy.types.Array2D` using `Axes.pcolormesh`."""
        x = numpy.concatenate((array.xindex.value, array.xspan[-1:]))
        y = numpy.concatenate((array.yindex.value, array.yspan[-1:]))
        xcoord, ycoord = numpy.meshgrid(x, y, copy=False, sparse=True)
        return self.pcolormesh(
            xcoord,
            ycoord,
            array.value.T,
            **kwargs,
        )

    def hist(
        self,
        x: ArrayLike | Sequence[ArrayLike],
        bins: int | Sequence[float] | str | None = None,
        **kwargs,
    ) -> tuple[
        numpy.ndarray | list[numpy.ndarray],
        numpy.ndarray,
        BarContainer | Polygon | list[BarContainer | Polygon],
    ]:
        """Draw a histogram of some data."""
        x = numpy.asarray(x)

        # re-format weights as array if given as float
        weights = kwargs.get("weights")
        if isinstance(weights, Number):
            kwargs["weights"] = numpy.ones_like(x) * weights

        # calculate log-spaced bins on-the-fly
        if (
            # note: this needs to be first to ensure pop()
            kwargs.pop("logbins", False)
            and isinstance(bins, int | None)
        ):
            nbins = int(bins or rcParams.get("hist.bins", 30))
            # get range
            hrange = kwargs.pop("range", None)
            if hrange is None:
                try:
                    hrange = numpy.min(x), numpy.max(x)
                except ValueError as exc:
                    if str(exc).startswith("zero-size array"):  # no data
                        exc.args = (
                            "cannot generate log-spaced histogram bins for zero-"
                            "size array, please pass `bins` or `range` manually",
                        )
                    raise
            # log-scale the axis and extract the base
            if kwargs.get("orientation") == "horizontal":
                self.set_yscale("log", nonpositive="clip")
                logbase = self.yaxis._scale.base
            else:
                self.set_xscale("log", nonpositive="clip")
                logbase = self.xaxis._scale.base
            # generate the bins
            bins = numpy.logspace(
                log(hrange[0], logbase),
                log(hrange[1], logbase),
                nbins + 1,
                endpoint=True,
            )

        return super().hist(x, bins=bins, **kwargs)

    hist.__doc__ = _Axes.hist.__doc__.replace(  # type: ignore[union-attr]
        "color :",
        "logbins : boolean, optional\n"
        "    If ``True``, use logarithmically-spaced histogram bins.\n\n"
        "    Default is ``False``\n\n"
        "color :")

    # -- new plotting methods --------

    def plot_mmm(
        self,
        data: Series,
        lower: Series | None = None,
        upper: Series | None = None,
        **kwargs,
    ) -> list[Artist]:
        """Plot a `Series` as a line, with a shaded region around it.

        The ``data`` `Series` is drawn, while the ``lower`` and ``upper``
        `Series` are plotted lightly below and above, with a fill
        between them and the ``data``.

        All three `Series` should have the same `~Series.index` array.

        Parameters
        ----------
        data : `~gwpy.types.Series`
            Data to plot normally.

        lower : `~gwpy.types.Series`
            Lower boundary (on Y-axis) for shade.

        upper : `~gwpy.types.Series`
            Upper boundary (on Y-axis) for shade.

        **kwargs
            Any other keyword arguments acceptable for
            :meth:`~matplotlib.Axes.plot`.

        Returns
        -------
        artists : `tuple`
            All of the drawn artists:

            - `~matplotlib.lines.Line2d` for ``data``,
            - `~matplotlib.lines.Line2D` for ``lower``, if given
            - `~matplotlib.lines.Line2D` for ``upper``, if given
            - `~matplitlib.collections.PolyCollection` for shading

        See Also
        --------
        matplotlib.axes.Axes.plot
            for a full description of acceptable ``*args`` and ``**kwargs``
        """
        alpha = kwargs.pop("alpha", .1)

        # plot mean
        line, = self.plot(data, **kwargs)
        out: list[Artist] = [line]

        # modify keywords for shading
        kwargs.update({
            "label": "",
            "linewidth": line.get_linewidth() / 2,
            "color": line.get_color(),
            "alpha": alpha * 2,
        })

        # plot lower and upper Series
        fill = [data.xindex.value, data.value, data.value]
        for i, bound in enumerate((lower, upper)):
            if bound is not None:
                out.extend(self.plot(bound, **kwargs))
                fill[i+1] = bound.value

        # fill between
        out.append(self.fill_between(
            *fill,
            alpha=alpha,
            color=kwargs["color"],
            rasterized=kwargs.get("rasterized", True),
        ))

        return out

    def tile(
        self,
        x: ArrayLike,
        y: ArrayLike,
        w: ArrayLike,
        h: ArrayLike,
        color: ArrayLike | None = None,
        anchor: Literal["center", "ll", "lr", "ul", "ur"] = "center",
        edgecolors: ColorType | Sequence[ColorType] | None = "face",
        linewidth: float | Sequence[float] | None = 0.8,
        **kwargs,
    ) -> PolyCollection:
        """Plot rectanguler tiles based onto these `Axes`.

        ``x`` and ``y`` give the anchor point for each tile, with
        ``w`` and ``h`` giving the extent in the X and Y axis respectively.

        Parameters
        ----------
        x, y, w, h : `array_like`, shape (n, )
            Input data

        color : `array_like`, shape (n, )
            Array of amplitudes for tile color

        anchor : `str`, optional
            Anchor point for tiles relative to ``(x, y)`` coordinates, one of

            - ``'center'`` - center tile on ``(x, y)``
            - ``'ll'`` - ``(x, y)`` defines lower-left corner of tile
            - ``'lr'`` - ``(x, y)`` defines lower-right corner of tile
            - ``'ul'`` - ``(x, y)`` defines upper-left corner of tile
            - ``'ur'`` - ``(x, y)`` defines upper-right corner of tile

        edgecolors : colour or list of colors, optional
            Edge colour for each tile.
            Default is the special value ``"face"`` which matches the edgecolor
            to the facecolor.

        linewidth : `float`, array of `float`, optional
            Line width for each tile.

        **kwargs
            Other keywords are passed to
            :meth:`~matplotlib.collections.PolyCollection`.

        Returns
        -------
        collection : `~matplotlib.collections.PolyCollection`
            The collection of tiles drawn.

        Examples
        --------
        >>> import numpy
        >>> from matplotlib import pyplot
        >>> import gwpy.plot  # to get gwpy's Axes

        >>> x = numpy.arange(10)
        >>> y = numpy.arange(x.size)
        >>> w = numpy.ones_like(x) * .8
        >>> h = numpy.ones_like(x) * .8

        >>> fig = pyplot.figure()
        >>> ax = fig.gca()
        >>> ax.tile(x, y, w, h, anchor='ll')
        >>> pyplot.show()
        """
        if kwargs.pop("sortbycolor", False) and color is not None:
            # try and sort the colour array by value
            try:
                color, x, y, w, h = _sortby(color, x, y, w, h)
            except ValueError as exc:
                exc.args = (
                    "Axes.tile argument 'sortbycolor' can only be used "
                    "with a simple array of floats in the `color` array",
                )
                raise

        # define how to make a polygon for each tile
        try:
            _poly = globals()[f"_poly_{anchor}"]
        except KeyError:
            msg = f"unrecognised tile anchor '{anchor}'"
            raise ValueError(msg) from None

        # build collection
        cmap = kwargs.pop("cmap", rcParams["image.cmap"])
        coll = PolyCollection(
            numpy.fromiter(
                (
                    _poly(*tile) for tile in
                    zip(x, y, w, h, strict=True)  # type: ignore[arg-type]
                ),
                dtype=(float, (4, 2)),
                count=numpy.shape(x)[0],
            ),
            edgecolors=edgecolors,
            linewidth=linewidth,
            **kwargs,
        )
        if color is not None:
            coll.set_array(color)
            coll.set_cmap(cmap)

        self.add_collection(coll)
        self.autoscale_view()
        return coll

    # -- overloaded auxiliary methods

    def legend(
        self,
        *args,  # noqa: ANN002
        **kwargs,
    ) -> Legend:
        """Generate a `~matplotlib.legend.Legend` for these `Axes`.

        This custom method just inserts our custom `HandlerLine2D` handler into the
        default ``handler_map``.
        """
        # build custom handler to render thick lines by default
        handler_map = kwargs.setdefault("handler_map", {})
        if isinstance(handler_map, dict):
            handler_map.setdefault(Line2D, HandlerLine2D(6))

        # create legend
        return super().legend(*args, **kwargs)

    legend.__doc__ = _Axes.legend.__doc__.replace(  # type: ignore[union-attr]
        "Call signatures",
        """.. note::

   This method uses a custom default legend handler for
   `~matplotlib.lines.Line2D` objects, with increased linewidth relative
   to the upstream :meth:`~matplotlib.axes.Axes.legend` method.
   To disable this, pass ``handler_map=None``, or create and pass your
   own handler class.  See :ref:`gwpy-plot-legend` for more details.

Call signatures""",
    )

    def colorbar(
        self,
        mappable: AxesImage | Collection | None = None,
        fraction: float = 0.,
        **kwargs,
    ) -> Colorbar:
        """Add a `~matplotlib.colorbar.Colorbar` to these `Axes`.

        Parameters
        ----------
        mappable : matplotlib data collection, optional
            Collection against which to map the colouring, default will
            be the last added mappable artist (collection or image).

        fraction : `float`, optional
            Fraction of space to steal from these `Axes` to make space
            for the new axes, default is ``0.``.
            Use ``fraction=.15`` to match the upstream matplotlib default.

        **kwargs
            other keyword arguments to be passed to the
            :meth:`Plot.colorbar` generator

        Returns
        -------
        cbar : `~matplotlib.colorbar.Colorbar`
            the newly added `Colorbar`

        See Also
        --------
        Plot.colorbar
        """
        return colorbar(
            self.get_figure(),
            ax=self,
            mappable=mappable,
            fraction=fraction,
            **kwargs,
        )


# override default Axes with this one by registering a projection with the
# same name

register_projection(Axes)


# -- overload Axes.plot() to handle Series

class PlotArgsProcessor(_process_plot_var_args):
    """`Axes.plot()` argument processor."""

    def __call__(
        self,
        *args: float | ArrayLike,
        **kwargs,
    ) -> list[Artist]:
        """Find `Series` data in `plot()` args and unwrap."""
        from gwpy.types import Series

        newargs: list[float | ArrayLike] = []

        # matplotlib 3.8.0 includes the Axes object up-front
        if args and isinstance(args[0], Axes):
            newargs.append(args[0])
            args = args[1:]

        while args:
            # strip first argument
            this, args = args[:1], args[1:]
            # it its a 1-D Series, then parse it as (xindex, value)
            if isinstance(this[0], Series) and this[0].ndim == 1:
                this = (this[0].xindex.value, this[0].value)
            # otherwise treat as normal (must be a second argument)
            else:
                this += args[:1]
                args = args[1:]
            # allow colour specs
            if args and isinstance(args[0], str):
                this += (args[0],)
                args = args[1:]
            newargs.extend(this)

        return super().__call__(*newargs, **kwargs)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.types.index`."""

from astropy import units

from .. import Index


class TestIndex:
    """Test `gwpy.types.Index`."""

    TEST_CLASS = Index

    def test_define_regular(self):
        """Check for regression against gwpy/gwpy#1596."""
        ind_len = 1000
        a = self.TEST_CLASS.define(
            1262936373.4853957,
            0.051,
            ind_len,
        )
        assert len(a) == ind_len

    def test_define_index_length(self):
        """Check for regression against gwpy/gwpy#1506."""
        a = self.TEST_CLASS.define(
            units.Quantity(1000000000),
            units.Quantity(0.01),
            500,
        )
        assert a.is_regular()

    def test_is_regular(self):
        """Test `Index.is_regular()`."""
        a = self.TEST_CLASS([1, 2, 3, 4, 5, 6], "s")
        assert a.is_regular()
        assert a[::-1].is_regular()

    def test_is_regular_scalar(self):
        """Test `Index.is_regular()` returns `False` for scalars."""
        a = self.TEST_CLASS([1])
        assert not a.is_regular()

    def test_not_is_regular(self):
        """Test `Index.is_regular()` (return `False`)."""
        b = self.TEST_CLASS([1, 2, 4, 5, 7, 8, 9])
        assert not b.is_regular()

    def test_regular(self):
        """Test `Index.regular`."""
        a = self.TEST_CLASS([1, 2, 3, 4, 5, 6], "s")
        assert a.regular
        assert a.regular is a.info.meta["regular"]

    def test_getitem(self):
        """Test `Index.__getitem__()`."""
        a = self.TEST_CLASS([1, 2, 3, 4, 5, 6], "Hz")
        assert type(a[0]) is units.Quantity
        assert a[0] == 1 * units.Hz
        assert isinstance(a[:2], type(a))

    def test_setitem(self):
        """Test `Index.__setitem__()`."""
        a = self.TEST_CLASS([1, 2, 3, 4, 5, 6], "Hz")
        a[0:2] = [7, 8] * a.unit
        assert list(a.value) == [7, 8, 3, 4, 5, 6]
        assert a[0] == 7 * units.Hz



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Logging utilities for GWpy."""

from __future__ import annotations

import logging
import os
import sys
from typing import TYPE_CHECKING

try:
    from coloredlogs import (
        ColoredFormatter,
        terminal_supports_colors as _terminal_supports_colors,
    )
except ImportError:
    def _terminal_supports_colors(stream: IO) -> bool:   # noqa: ARG001
        """Return `False` to indicate that colours are unsupported."""
        return False

if TYPE_CHECKING:
    from typing import IO

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = [
    "DEFAULT_LOG_DATEFMT",
    "DEFAULT_LOG_FORMAT",
    "get_default_level",
    "init_logger",
]

_LOG_LEVELS = logging.getLevelNamesMapping()

#: The default log message format.
#:
#: The format can be overridden by setting the ``GWPY_LOG_FORMAT``
#: environment variable.
DEFAULT_LOG_FORMAT = os.getenv(
    "GWPY_LOG_FORMAT",
    "%(asctime)s:%(name)s:%(levelname)s:%(message)s",
)

#: The default log date format.
#:
#: The format can be overridden by setting the ``GWPY_LOG_DATEFMT``
#: environment variable.
DEFAULT_LOG_DATEFMT = os.getenv(
    "GWPY_LOG_DATEFMT",
    "%Y-%m-%dT%H:%M:%S%z",
)


def get_default_level() -> int:
    """Return the default log level by inspecting the ``GWPY_LOG_LEVEL`` variable.

    If ``GWPY_LOG_LEVEL`` is not set, `logging.NOTSET` is returned.

    Returns
    -------
    level : `int`
        The log level to use by default for loggers.

    Examples
    --------
    >>> os.environ["GWPY_LOG_LEVEL"] = "debug"
    >>> get_default_level()
    10
    >>> os.environ["GWPY_LOG_LEVEL"] = logging.INFO
    >>> get_default_level()
    20
    """
    try:
        level = os.environ["GWPY_LOG_LEVEL"].upper()
    except KeyError:
        return logging.NOTSET
    if level.isdigit():
        return int(level)
    return _LOG_LEVELS[level]


def init_logger(
    name: str,
    level: int | str | None = None,
    *,
    stream: IO = sys.stderr,
    fmt: str = DEFAULT_LOG_FORMAT,
    datefmt: str | None = DEFAULT_LOG_DATEFMT,
    style: str = "%",
    color: bool = True,
) -> logging.Logger:
    """Return the logger with the given name, or create one as needed.

    This function calls `logging.getLogger` to find a logger with the
    specified ``name``, or to create one.

    If the returned logger has no handlers attached already, a new
    `logging.StreamHandler` will be created and attached based
    on ``stream``.

    Parameters
    ----------
    name : `str`
        The name of the logger.

    level : `int`, `str`, optional
        The level to set on the logger.
        If ``level=None`` (default) is given, the logging level will be
        determined from the ``GWPY_LOG_LEVEL`` environment variable, or
        set to `logging.NOTSET`.

    stream : `io.IOBase`, optional
        The stream to write log messages to.

    fmt : `str`, optional
        The message format to use for a new handler.

    datefmt : `str`, optional
        The date format to use for a new handler.

    style : `str`, optional
        The type of ``format`` string.

    color : `bool`, optional
        If `True` (default) try to use `coloredlogs.ColoredFormatter`
        as the default formatter.
        If `False` the default formatter is `logging.Formatter`.

    Returns
    -------
    logger : `logging.Logger`
        A new, or existing, `Logger` instance, with at least one
        `~logging.Handler` attached.

    Notes
    -----
    If ``color=True`` is given (default) the
    :func:`~humanfriendly.terminal.terminal_supports_colors` function is
    called to determine whether the output terminal can support colours.
    If that returns `False`, coloured formatting is not used.

    See Also
    --------
    logging.basicConfig
        For more complete descriptions of some of the keyword arguments.
    """
    # get the default level
    if level is None:
        level = get_default_level()

    # get the logger
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # attach a new handler as needed
    if not logger.hasHandlers():
        handler = logging.StreamHandler(stream)
        if color and _terminal_supports_colors(stream):
            formatter_class = ColoredFormatter
        else:
            formatter_class = logging.Formatter
        handler.setFormatter(formatter_class(
            fmt=fmt,
            datefmt=datefmt,
            style=style,
        ))
        logger.addHandler(handler)
    return logger



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read a `Series` from an ASCII file.

These files should be in (at least) two-column x,y format.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from numpy import (
    column_stack,
    loadtxt,
    savetxt,
)

from ...io.registry import identify_factory
from .. import (
    Array2D,
    Series,
)

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path
    from typing import IO

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- read ----------------------------

def read_ascii_series(
    input_: str | Path | IO,
    array_type: type[Series] = Series,
    *,
    unpack: bool = True,
    **kwargs,
) -> Series:
    """Read a `Series` from an ASCII file.

    Parameters
    ----------
    input_ : `str`, `file`
        File to read.

    array_type : `type`
        Desired return type.

    unpack : `bool`
        If `True` (default), unpack the data into separate columns.
        This argument is passed to `numpy.loadtxt`.

    kwargs
        Additional keyword arguments to pass to `numpy.loadtxt`.

    Returns
    -------
    Series
        The data read from the file, as a `Series` object.

    See Also
    --------
    numpy.loadtxt
        For documentation of keyword arguments.
    """
    # load data
    data = loadtxt(input_, unpack=unpack, **kwargs)
    # separate first column as index
    xindex = data[0]
    # reshape other columns as data
    if array_type._ndim == 1:
        ydata = data[1:][0]  # one column
    else:
        ydata = data[1:].T  # ND array
    return array_type(ydata, xindex=xindex)


# -- write ---------------------------

def write_ascii_series(
    series: Series,
    output: str | Path | IO,
    **kwargs,
) -> None:
    """Write a `Series` to a file in ASCII format.

    Parameters
    ----------
    series : `Series`
        Data series to write.

    output : `str`, `file`
        File to write to.

    kwargs
        Additional keyword arguments to pass to `numpy.savetxt`.

    See Also
    --------
    numpy.savetxt
        For documentation of keyword arguments.
    """
    xarr = series.xindex.value
    yarr = series.value
    savetxt(output, column_stack((xarr, yarr)), **kwargs)


# -- register ------------------------


def register_ascii_io(
    array_type: type[Series],
    format: str = "txt",  # noqa: A002
    *,
    identify: bool = True,
    reader: Callable = read_ascii_series,
    writer: Callable = write_ascii_series,
    **defaults,
) -> None:
    """Register ASCII read/write/identify methods for the given type."""

    def _read(source: str | Path | IO, **kwargs) -> Series:
        kwgs = defaults.copy()
        kwgs.update(kwargs)
        return reader(
            source,
            array_type=array_type,
            **kwgs,
        )

    def _write(
        series: Series,
        output: str | Path | IO,
        **kwargs,
    ) -> None:
        kwgs = defaults.copy()
        kwgs.update(kwargs)
        return writer(series, output, **kwgs)

    array_type.read.registry.register_reader(format, array_type, _read)
    array_type.write.registry.register_writer(format, array_type, _write)
    if identify:
        array_type.read.registry.register_identifier(
            format,
            array_type,
            identify_factory(format),
        )


register_ascii_io(Series, "csv", delimiter=",")
register_ascii_io(Array2D, "csv", delimiter=",")



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Methods for sensitivity calculations of gravitational-wave interferometer data.

The LIGO project measures time-dependent sensitivity by calculating the
distance at which the gravitational-wave signature of a binary neutron star
(BNS) inspiral would be recorded by an instrument with a signal-to-noise ratio
(SNR) of 8.
In most of the literature, this is known as the 'inspiral range' or the
'horizon distance'.
"""

from .range import (
    burst_range,
    burst_range_spectrum,
    inspiral_range,
    inspiral_range_psd,
    range_spectrogram,
    range_timeseries,
    sensemon_range,
    sensemon_range_psd,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"



# Copyright (c) 2017-2020 Scott Coughlin
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for database queries.

All functions in this module require :doc:`sqlalchemy <sqlalchemy:index>`.
"""

from __future__ import annotations

import operator
from functools import reduce
from typing import (
    TYPE_CHECKING,
    cast,
)

from astropy.table import Table

from .. import EventTable
from ..filter import parse_column_filters

if TYPE_CHECKING:
    from collections.abc import (
        Iterable,
        Mapping,
        Sequence,
    )
    from typing import (
        Any,
        TypeAlias,
    )

    import sqlalchemy

    WhereExpression: TypeAlias = sqlalchemy.SQLColumnExpression | str | None

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- utilities -----------------------

def format_where(
    condition: WhereExpression | Iterable[WhereExpression],
) -> sqlalchemy.SQLColumnExpression | None:
    """Format a column filter condition as a SQL ``WHERE`` expression.

    Requires: :doc:`sqlalchemy <sqlalchemy:index>`

    Parameters
    ----------
    condition : `str`, `list` of `str`
        A column filter string or list of filters.
        Each entry should be parseable using
        `gwpy.table.filter.parse_column_filter`.

    Returns
    -------
    expression : `sqlalchemy.SQLColumnExpression`, `None`
        The formatted query expression, suitable for application using
        `~sqlalchemy.sql.expression.Select.where`.
        `None` will be returned if no ``WHERE`` expression is created
        (``condition`` is `None` or an empty iterable).

    Examples
    --------
    >>> format_where(['snr > 10', 'frequency < 1000'])
    <sqlalchemy.sql.elements.BooleanClauseList object at 0x7f0d7faf78f0>
    >>> print(format_where(['snr > 10', 'frequency < 1000']))
    snr > :snr_1 AND frequency < :frequency_1
    """
    from sqlalchemy import (
        Column,
        SQLColumnExpression,
    )

    # parse condition for SQL query
    if condition is None:
        return None

    if isinstance(condition, str | SQLColumnExpression):
        condition = [condition]

    expressions = []
    for item in condition:
        if item is None:
            continue
        if isinstance(item, SQLColumnExpression):
            expressions.append(item)
        else:
            for col, op_, value in parse_column_filters(item):
                val = cast("SQLColumnExpression", op_(Column(col), value))
                expressions.append(val)
    if expressions:
        return reduce(operator.and_, expressions)
    return None


def format_query(
    tablename: str | sqlalchemy.TableClause,
    columns: Iterable[str | sqlalchemy.Column] | None = None,
    *,
    where: str | list[str] | None = None,
    order_by: str | sqlalchemy.Column | None = None,
    order_by_desc: bool = False,
) -> sqlalchemy.Select:
    """Format a SQL query using `sqlalchemy`.

    Requires: :doc:`sqlalchemy <sqlalchemy:index>`

    Parameters
    ----------
    tablename: `str`
        The name of the database table to ``SELECT FROM``.

    columns : `list` of `str`, optional
        The list of columns to ``SELECT``.

    where : `str`, `list` of `str`, optional
        A filter or list of filters to apply as ``WHERE`` conditions.

    order_by : `str`, optional
        The column to ``ORDER BY``.

    order_by_desc : `bool`, optional
        If `True`, apply ``order_by`` with the ``DESC`` flag.
        Default is `False` (``ASC``).

    Returns
    -------
    statement : `sqlalchemy.sql.expression.Select`
        The formatted ``SELECT`` statement, suitable for passing to
        `sqlalchemy.engine.Connection.execute` or `pandas.read_sql`.
    """
    from sqlalchemy import (
        Column,
        asc,
        desc,
        select,
        table,
    )

    # parse columns
    if columns is None:
        columns = ["*"]
    else:
        columns = list(map(Column, columns))
    columns = cast("list[sqlalchemy.Column]", columns)

    # build SQL query
    if isinstance(tablename, str):
        tablename = table(tablename)
    query = select(*columns).select_from(tablename)
    if (expr := format_where(where)) is not None:
        query = query.where(expr)
    if order_by:
        if order_by_desc:
            return query.order_by(desc(Column(order_by)))
        return query.order_by(asc(Column(order_by)))
    return query


# -- misc queries --------------------

def get_columns(
    tablename: str,
    engine: sqlalchemy.Engine,
) -> dict[str, type]:
    """Return the names and types of the columns in a database table.

    This is just a wrapper around
    `sqlalchemy.engine.reflection.Inspector.get_columns`.

    Parameters
    ----------
    tablename : `str`
        The name of table to inspect.

    engine : `sqlalchemy.engine.Engine`, optional
        The database engine to use when connecting.

    Returns
    -------
    names : `dict` of (`str`, `type`) pairs
        The name and type of each column found in the database.
    """
    from sqlalchemy import inspect

    inspector = inspect(engine)
    return {
        col["name"]: col["type"].python_type
        for col in inspector.get_columns(tablename)
    }


# -- fetch ---------------------------

def create_engine(
    drivername: str,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    database: str | None = None,
    query: Mapping[str, Sequence[str] | str] | None = None,
    **kwargs,
) -> sqlalchemy.Engine:
    """Create a new `sqlalchemy.engine.Engine`.

    Requires: :doc:`sqlalchemy <sqlalchemy:index>`

    Parameters
    ----------
    drivername : `str`
        Database backend and driver name.

    username : `str`, optional
        The username for authentication to the database.

    password : `str`, optional
        The password for authentication to the database.

    host : `str`, optional
        The name of the remote database host.

    port : `int`, optional
        Port to connect to on ``host``.

    database : `str`, optional
        The name of the database to connect to.

    query : `dict`, optional
        Query parameters.

    kwargs
        Other keyword arguments are passed to `sqlalchemy.create_engine`.

    Returns
    -------
    engine : `sqlalchemy.engine.Engine`
        A new engine.

    See Also
    --------
    sqlalchemy.engine.URL.create
        For details of how URLs are constructed.
    sqlalchemy.create_engine
        For documentation on the engine.
    """
    from sqlalchemy import (
        URL,
        create_engine,
    )

    url = URL.create(
        drivername,
        username=username,
        password=password,
        host=host,
        port=port,
        database=database,
        query=query or {},
    )
    return create_engine(url, **kwargs)


def fetch(
    # query options
    tablename: str | sqlalchemy.TableClause,
    columns: Iterable[str | sqlalchemy.Column] | None = None,
    *,
    where: str | list[str] | None = None,
    order_by: str | None = None,
    order_by_desc: bool = False,
    # connection options
    engine: sqlalchemy.Engine | None = None,
    drivername: str | None = None,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    database: str | None = None,
    query: dict[str, Any] | None = None,
    **kwargs,
) -> Table:
    """Fetch data from an SQL table into a `Table`.

    Requires: :doc:`sqlalchemy <sqlalchemy:index>`

    Parameters
    ----------
    tablename : `str`
        The name of table you are attempting to receive triggers
        from.

    columns : `list` of `str`, optional
        The list of columns to ``SELECT``.

    where : `str`, `list` of `str`, optional
        A filter or list of filters to apply as ``WHERE`` conditions,
        e.g. ``'snr > 5'``.
        Multiple conditions should be connected by ' && ', or given as
        a `list`, e.g. ``'snr > 5 && frequency < 1000'`` or
        ``['snr > 5', 'frequency < 1000']``

    order_by : `str`, optional
        The column to ``ORDER BY``.

    order_by_desc : `bool`, optional
        If `True`, apply ``order_by`` with the ``DESC`` flag.
        Default is `False` (``ASC``).

    engine : `sqlalchemy.engine.Engine`, optional
        The database engine to use when connecting.

    drivername : `str`, optional
        Database backend and driver name.
        This is required if ``engine`` is not specified.

    username : `str`, optional
        The username for authentication to the database.

    password : `str`, optional
        The password for authentication to the database.

    host : `str`, optional
        The name of the remote database host.

    port : `int`, optional
        Port to connect to on ``host``.

    database : `str`, optional
        The name of the database to connect to.

    query : `dict`, optional
        Query parameters.

    kwargs
        Other keyword arguments are passed to `pandas.read_sql`.

    Returns
    -------
    table : `Table`
        The table of data received from the SQL query.
    """
    import pandas

    # create engine
    if engine is None:
        if drivername is None:
            msg = "either 'engine' or 'drivername' must be specified"
            raise ValueError(msg)
        engine = create_engine(
            drivername=drivername,
            username=username,
            password=password,
            host=host,
            port=port,
            database=database,
            query=query,
        )

    # build query
    qstr = format_query(
        tablename,
        columns=columns,
        where=where,
        order_by=order_by,
        order_by_desc=order_by_desc,
    )

    # kwargs for Table.from_pandas
    index = kwargs.pop("index", False)
    units = kwargs.pop("units", None)

    # perform query
    dataframe = pandas.read_sql(
        qstr,
        engine,
        **kwargs,
    )

    # Convert unicode columns to string
    types = dataframe.apply(lambda x: pandas.api.types.infer_dtype(x.values))
    if not dataframe.empty:
        for col in types[types == "unicode"].index:
            dataframe[col] = dataframe[col].astype(str)

    return Table.from_pandas(
        dataframe,
        index=index,
        units=units,
    ).filled()


EventTable.fetch.registry.register_reader(
    "sql",
    EventTable,
    fetch,
)



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unified I/O for :mod:`gwpy.timeseries` objects."""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from ..io import cache as io_cache
from ..io.registry import (
    UnifiedGet,
    UnifiedRead,
)
from ..time import to_gps
from ..types.connect import SeriesWrite

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import (
        Any,
        Literal,
        SupportsFloat,
        TypeVar,
    )

    from numpy.typing import DTypeLike

    from ..detector import Channel
    from ..io.utils import NamedReadable
    from ..time import SupportsToGps
    from . import (
        TimeSeriesBase,
        TimeSeriesBaseDict,
    )

    T = TypeVar("T", bound=TimeSeriesBase | TimeSeriesBaseDict)
    TimeSeriesType = TypeVar("TimeSeriesType", bound=TimeSeriesBase)
    TimeSeriesDictType = TypeVar("TimeSeriesDictType", bound=TimeSeriesBaseDict)


# -- utilities -----------------------

def _pad_series(
    ts: TimeSeriesType,
    pad: float | None,
    start: SupportsFloat | None = None,
    end: SupportsFloat | None = None,
    *,
    error: bool = False,
) -> TimeSeriesType:
    """Pad a timeseries to match the specified [start, end) limits.

    To cover a gap in data returned from a data source.

    Parameters
    ----------
    ts : `gwpy.types.Series`
        The input series.

    pad : `float`, `astropy.units.Quantity`
        The value with which to pad.

    start : `float`, `astropy.units.Quantity`, optional
        The desired start point of the X-axis, defaults to
        the start point of the incoming series.

    end : `float`, `astropy.units.Quantity`, optional
        The desired end point of the X-axis, defaults to
        the end point of the incoming series.

    error : `bool`, optional
        Raise `ValueError` when gaps are present, rather than padding
        anything.

    Returns
    -------
    series : instance of incoming series type
        A padded version of the series. This may be the same
        object if not padding is needed.

    Raises
    ------
    ValueError
        If `error=True` is given and padding would have been required
        to match the request.
    """
    span = ts.span
    if start is None:
        start = span[0]
    if end is None:
        end = span[1]
    pada = max(int((span[0] - start) * ts.sample_rate.value), 0)
    padb = max(int((end - span[1]) * ts.sample_rate.value), 0)
    if not (pada or padb):  # if noop, just return the input
        return ts
    if error:  # if error, bail out now
        msg = (
            f"{type(ts).__name__} with span {span} does not cover "
            f"requested interval {type(span)(start, end)}"
        )
        raise ValueError(msg)
    # otherwise applying the padding
    return ts.pad((pada, padb), mode="constant", constant_values=(pad,))


# -- Unified I/O singletons ----------

class _TimeSeriesRead(UnifiedRead):
    """Base `TimeSeriesRead` implementation.

    This defines the logic, but all concrete instances should be created
    from a subclass that gives a detailed docstring to support
    `Klass.read.help()`.
    """

    @abstractmethod
    def merge(  # type: ignore[override]
        self,
        items: Sequence[T],
        pad: float | None = None,
        gap: Literal["raise", "ignore", "pad"] | None = None,
        start: SupportsFloat | None = None,
        end: SupportsFloat | None = None,
    ) -> T:
        """Combine a list of `TimeSeries` or `TimeSeriesDict` into one.

        Must be given at least one item.
        """

    def __call__(  # type: ignore[override]
        self,
        source: NamedReadable | Sequence[NamedReadable],
        name: str | Channel | None = None,
        start: SupportsToGps | None = None,
        end: SupportsToGps | None = None,
        *,
        pad: float | None = None,
        gap: Literal["raise", "ignore", "pad"] | None = None,
        **kwargs,
    ) -> TimeSeriesBase:
        """Read a `TimeSeries` from a source."""
        if start is not None:
            start = to_gps(start)
        if end is not None:
            end = to_gps(end)

        # if reading a cache, read it now and sieve
        if io_cache.is_cache(source):
            from .io.cache import preformat_cache
            source = preformat_cache(
                source,
                start=start,
                end=end,
            )

        # construct parametrised merge function
        # (this allows pad, gap, start, end to be passed into
        #  the underlying read function)
        if gap is None:
            gap = "raise" if pad is None else "pad"
        def merge(items: Sequence[T]) -> T:
            return self.merge(
                items,
                pad=pad,
                gap=gap,
                start=start,
                end=end,
            )

        # read (with optional name)
        args = () if name is None else (name,)
        return super().__call__(
            source,
            *args,
            start=start,
            end=end,
            merge_function=merge,
            **kwargs,
        )


# -- TimeSeriesBase -----------------

class TimeSeriesBaseRead(_TimeSeriesRead):
    """Read data into a `TimeSeriesBase`.

    Notes
    -----"""

    def merge(  # type: ignore[override]
        self,
        items: Sequence[TimeSeriesBase],
        pad: float | None = None,
        gap: Literal["raise", "ignore", "pad"] | None = None,
        start: SupportsFloat | None = None,
        end: SupportsFloat | None = None,
    ) -> TimeSeriesBase:
        """Combine a list of `TimeSeriesBase` objects into one `Series`."""
        from . import TimeSeriesBaseList
        if len(items) == 1:  # don't copy a single array
            joined = items[0]
        else:
            list_ = TimeSeriesBaseList(*items)
            joined = list_.join(pad=pad, gap=gap)
        if gap in ("pad", "raise"):
            return _pad_series(
                joined,
                pad,
                start,
                end,
                error=(gap == "raise"),
            )
        if not isinstance(joined, self._cls):
            joined = self._cls(joined)
        return joined


class TimeSeriesBaseWrite(SeriesWrite):
    """Write data from a `TimeSeriesBase`.

    Notes
    -----"""


class TimeSeriesBaseGet(UnifiedGet):
    """Get data from any source.

    Notes
    -----"""

    def __call__(  # type: ignore[override]
        self,
        name: str | Channel,
        start: SupportsToGps,
        end: SupportsToGps,
        *,
        source: str | list[str | dict[str, Any]] | None = None,
        **kwargs,
    ) -> TimeSeriesBase:
        """Retrieve data from a source."""
        dict_method = getattr(self._cls.DictClass, self.method)
        kwargs.setdefault("series_class", self._cls)
        return dict_method(
            [name],
            start,
            end,
            source=source,
            **kwargs,
        )[name]


# -- TimeSeries ---------------------

class TimeSeriesRead(TimeSeriesBaseRead):
    """Read data into a `TimeSeries`.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the parameters
    below are common to all formats.

    Parameters
    ----------
    source : `str`, `os.PathLike`, `file`, or `list` of these
        Source of data, any of the following:

        - Path of a single data file
        - List of data file paths
        - Path of LAL-format cache file

    name : `str`, optional
        The name of the data object to read, or a `Channel` object.
        This argument is required if ``source`` contains (or may contain)
        multiple datasets.

        - When reading from GWF, this argument should specify the
          name of the channel to read.
        - When reading from HDF5, this argument should specify the
          path or dataset name within the file.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS start time of required data, defaults to start of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    format : `str`, optional
        Source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    parallel : `int`, optional
        Number of parallel processes to use, serial process by
        default.

    pad : `float`, optional
        Value with which to fill gaps in the source data,
        by default gaps will result in a `ValueError`.

    Raises
    ------
    IndexError
        if ``source`` is an empty list

    Notes
    -----"""


class TimeSeriesWrite(TimeSeriesBaseWrite):
    """Write this `TimeSeries` to a file.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    target : `str`
        output filename

    format : `str`, optional
        output format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    Notes
    -----"""


class TimeSeriesGet(TimeSeriesBaseGet):
    """Retrieve data for a channel from any data source.

    This method attempts to get data any way it can, potentially iterating
    over multiple available data sources.

    Parameters
    ----------
    channels : `list`
        Required data channels.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS start time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS end time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine

    source : `str`, `list`, `list` of `dict`.
        The data source to use. One of the following formats:

        - `str` - the name of a single source to use,
        - `list` - a list of source names to try in order,
        - `list` of `dict` - a list of source specifications to try in order;
          each `dict` must contain a `"source"` key giving the name of the
          source to use, and may contain other keys giving options to
          pass to the data access function for that source.

        See 'Notes' section below for valid source names.

    frametype : `str`
        Name of frametype in which this channel is stored, by default
        will search for all required frame types.

    pad : `float`
        Value with which to fill gaps in the source data,
        by default gaps will result in a `ValueError`.

    scaled : `bool`
        apply slope and bias calibration to ADC data, for non-ADC data
        this option has no effect.

    nproc : `int`, default: `1`
        Number of parallel processes to use, serial process by
        default.

    allow_tape : `bool`, default: `None`
        Allow the use of data files that are held on tape.
        Default is `None` to attempt to allow the `TimeSeries.fetch`
        method to intelligently select a server that doesn't use tapes
        for data storage (doesn't always work), but to eventually allow
        retrieving data from tape if required.

    verbose : `bool`
        This argument is deprecated and will be removed in a future release.
        Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

    kwargs
        Other keyword arguments to pass to the data access function for
        each data source.

    Notes
    -----"""


# -- StateVector --------------------

class StateVectorRead(TimeSeriesBaseRead):
    """Read data into a `StateVector`.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the parameters
    below are common to most formats.

    Parameters
    ----------
    source : `str`, `os.PathLike`, `file`, or `list` of these
        Source of data, any of the following:

        - Path of a single data file
        - List of data file paths
        - Path of LAL-format cache file

    name : `str`, optional
        The name of the data object to read, or a `Channel` object.
        This argument is required if ``source`` contains (or may contain)
        multiple datasets.

        - When reading from GWF, this argument should specify the
          name of the channel to read.
        - When reading from HDF5, this argument should specify the
          path or dataset name within the file.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS start time of required data, defaults to start of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    bits : `list`, optional
        List of bits names for this `StateVector`, give `None` at
        any point in the list to mask that bit.

    format : `str`, optional
        Source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    parallel : `int`, optional
        Number of parallel processes to use, serial process by
        default.

    pad : `float`, optional
        Value with which to fill gaps in the source data,
        by default gaps will result in a `ValueError`.

    Raises
    ------
    IndexError
        If ``source`` is an empty list.

    Notes
    -----"""


class StateVectorWrite(SeriesWrite):
    """Write this `StateVector` to a file.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    target : `str`
        output filename

    format : `str`, optional
        output format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    Notes
    -----"""


class StateVectorGet(TimeSeriesBaseGet):
    """Retrieve `StateVector` data for a channel.

    This method attemps to get data any way it can, potentially iterating
    over multiple available data sources.

    Parameters
    ----------
    name : `str`, `~gwpy.detector.Channel`
        The name of the channel to read, or a `Channel` object.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS start time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS end time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine.

    bits : `Bits`, `list`, optional
        Definition of bits for this `StateVector`

    source : `str`, `list`, `list` of `dict`.
        The data source to use. One of the following formats:

        - `str` - the name of a single source to use,
        - `list` - a list of source names to try in order,
        - `list` of `dict` - a list of source specifications to try in order;
          each `dict` must contain a `"source"` key giving the name of the
          source to use, and may contain other keys giving options to
          pass to the data access function for that source.

        See 'Notes' section below for valid source names.

    nproc : `int`, optional
        Number of parallel processes to use, serial process by
        default.

    verbose : `bool`, optional
        This argument is deprecated and will be removed in a future release.
        Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

    kwargs
        Other keyword arguments to pass to the data access function for
        each data source.

    Notes
    -----"""


# -- TimeSeriesBaseDict -------------

class TimeSeriesBaseDictRead(_TimeSeriesRead):
    """Read data into a `TimeSeriesBaseDict`.

    Notes
    -----"""

    def merge(  # type: ignore[override]
        self,
        items: Sequence[TimeSeriesBaseDict],
        pad: float | None = None,
        gap: Literal["raise", "ignore", "pad"] | None = None,
        start: SupportsFloat | None = None,
        end: SupportsFloat | None = None,
    ) -> TimeSeriesBaseDict:
        """Combine a list of `TimeSeriesBaseDict` objects into one `Series`."""
        out = self._cls()
        for tsd in sorted(items, key=lambda x: x.span):
            out.append(tsd, gap=gap, pad=pad)
            del tsd
        if gap in ("pad", "raise"):
            for key in out:
                out[key] = _pad_series(
                    out[key],
                    pad,
                    start,
                    end,
                    error=(gap == "raise"),
                )
        return out


class TimeSeriesBaseDictWrite(SeriesWrite):
    """Write data from a `TimeSeriesBaseDict`.

    Notes
    -----"""


class TimeSeriesBaseDictGet(UnifiedGet):
    """Retrieve data for multiple names from any data source.

    Notes
    -----"""

    def __call__(  # type: ignore[override]
        self,
        names: Sequence[str | Channel],
        start: SupportsToGps,
        end: SupportsToGps,
        *,
        source: str | list[str | dict[str, Any]] | None = None,
        pad: float | None = None,
        scaled: bool | None = None,
        dtype: DTypeLike | None = None,
        verbose: bool | None = None,
        allow_tape: bool | None = None,
        **kwargs,
    ) -> TimeSeriesBaseDict:
        """Retrieve data for multiple channels from any data source."""
        dictclass = self._cls
        entryclass = dictclass.EntryClass
        kwargs.setdefault("series_class", entryclass)
        try:
            return super().__call__(
                names,
                to_gps(start),
                to_gps(end),
                source=source,
                pad=pad,
                scaled=scaled,
                dtype=dtype,
                verbose=verbose,
                allow_tape=allow_tape,
                **kwargs,
            )
        except RuntimeError:
            # if we got here then we failed to get all data at once
            if len(names) == 1:
                raise
            self.logger.info(
                "Failed to access data for all names as a group, "
                "trying individually",
            )
            return dictclass((name, entryclass.get(
                name,
                start,
                end,
                source=source,
                pad=pad,
                scaled=scaled,
                dtype=dtype,
                verbose=verbose,
                allow_tape=allow_tape,
                **kwargs,
            )) for name in names)


class TimeSeriesDictGet(TimeSeriesBaseDictGet):
    """Retrieve data for multiple names from any data source.

    This method attemps to get data any way it can, potentially iterating
    over multiple available data sources.

    Parameters
    ----------
    names : `list`
        A `list` of channel names to find.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS start time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS end time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine

    source : `str`, `list`, `list` of `dict`.
        The data source to use. One of the following formats:

        - `str` - the name of a single source to use,
        - `list` - a list of source names to try in order,
        - `list` of `dict` - a list of source specifications to try in order;
          each `dict` must contain a `"source"` key giving the name of the
          source to use, and may contain other keys giving options to
          pass to the data access function for that source.

        See 'Notes' section below for valid source names.

    frametype : `str`
        Name of frametype in which this channel is stored, by default
        will search for all required frame types.

    pad : `float`
        Value with which to fill gaps in the source data,
        by default gaps will result in a `ValueError`.

    scaled : `bool`
        apply slope and bias calibration to ADC data, for non-ADC data
        this option has no effect.

    nproc : `int`, default: `1`
        Number of parallel processes to use, serial process by
        default.

    allow_tape : `bool`, default: `None`
        Allow the use of data files that are held on tape.
        Default is `None` to attempt to allow the `TimeSeries.fetch`
        method to intelligently select a server that doesn't use tapes
        for data storage (doesn't always work), but to eventually allow
        retrieving data from tape if required.

    verbose : `bool`
        This argument is deprecated and will be removed in a future release.
        Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

    kwargs
        Other keyword arguments to pass to the data access function for
        each data source. See `TimeSeriesDict.get.help(source=<>)` for details
        on the positional and keyword arguments supported for each data source.

    Notes
    -----"""


# -- TimeSeriesDict -----------------

class TimeSeriesDictRead(TimeSeriesBaseDictRead):
    """Read data into a `TimeSeriesDict`.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the parameters
    below are common to most formats.

    Parameters
    ----------
    source : `str`, `list`
        Source of data, any of the following:

        - `str` path of single data file,
        - `str` path of LAL-format cache file,
        - `list` of paths.

    names : Sequence of `str` or `~gwpy.detector.Channel`
        A list (or similar) of dataset names to read from ``source``.

        - When reading from GWF, this argument should specify the
          names of the channels to read.
        - When reading from HDF5, this argument should specify the
          paths or dataset names within the file.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS start time of required data, defaults to start of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    format : `str`, optional
        Source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    parallel : `int`, optional
        Number of parallel processes to use, serial process by
        default.

    pad : `float`, optional
        Value with which to fill gaps in the source data,
        by default gaps will result in a `ValueError`.

    Raises
    ------
    IndexError
        If ``source`` is an empty list.

    Notes
    -----"""


class TimeSeriesDictWrite(SeriesWrite):
    """Write this `TimeSeriesDict` to a file.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    target : `str`
        output filename

    format : `str`, optional
        output format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    Notes
    -----"""


# -- StateVectorDict ----------------

class StateVectorDictRead(TimeSeriesBaseDictRead):
    """Read data into a `StateVectorDict`.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the parameters
    below are common to most formats.

    Parameters
    ----------
    source : `str`, `list`
        Source of data, any of the following:

        - `str` path of single data file,
        - `str` path of LAL-format cache file,
        - `list` of paths.

    names : Sequence of `str` or `~gwpy.detector.Channel`
        A list (or similar) of dataset names to read from ``source``.

        - When reading from GWF, this argument should specify the
          names of the channels to read.
        - When reading from HDF5, this argument should specify the
          paths or dataset names within the file.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS start time of required data, defaults to start of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    format : `str`, optional
        Source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    parallel : `int`, optional
        Number of parallel processes to use, serial process by
        default.

    pad : `float`, optional
        Value with which to fill gaps in the source data,
        by default gaps will result in a `ValueError`.

    gap : `str`, optional
        How to handle gaps in the data, one of

        "ignore"
            Do nothing, let the underlying reader method handle it.

        "warn"
            Do nothing except print a warning to the screen.

        "raise"
            Raise an exception upon finding a gap (default).

        "pad"
            Insert a value to fill the gaps.

    Raises
    ------
    IndexError
        If ``source`` is an empty list.

    Notes
    -----"""


class StateVectorDictWrite(SeriesWrite):
    """Write this `StateVectorDict` to a file.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    target : `str`
        output filename

    format : `str`, optional
        output format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    Notes
    -----"""


class StateVectorDictGet(TimeSeriesBaseDictGet):
    """Retrieve `StateVector` data for multiple channels.

    This method attemps to get data any way it can, potentially iterating
    over multiple available data sources.

    Parameters
    ----------
    channels : `list`
        Required data channels.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS start time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS end time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine

    source : `str`, `list`, `list` of `dict`.
        The data source to use. One of the following formats:

        - `str` - the name of a single source to use,
        - `list` - a list of source names to try in order,
        - `list` of `dict` - a list of source specifications to try in order;
          each `dict` must contain a `"source"` key giving the name of the
          source to use, and may contain other keys giving options to
          pass to the data access function for that source.

        See 'Notes' section below for valid source names.

    kwargs
        Other keyword arguments to pass to the data access function for
        each data source. See `StateVectorDict.get.help(source=<>)` for details
        on the positional and keyword arguments supported for each data source.

    Notes
    -----"""



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for FFT routines."""

from astropy import units

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def scale_timeseries_unit(
    tsunit: units.UnitBase,
    scaling: str = "density",
) -> units.UnitBase:
    """Scale the unit of a `TimeSeries` to match that of a `FrequencySeries`.

    Parameters
    ----------
    tsunit : `~astropy.units.UnitBase`
        Input unit from `TimeSeries`.

    scaling : `str`
        Type of frequency series, either ``"density"`` for a PSD, or
        ``"spectrum"`` for a power spectrum.

    Returns
    -------
    unit : `~astropy.units.Unit`
        Unit to be applied to the resulting `FrequencySeries`.
    """
    # set units
    if scaling == "density":
        baseunit = units.Hertz
    elif scaling == "spectrum":
        baseunit = units.dimensionless_unscaled
    else:
        msg = f"unknown scaling: '{scaling}'"
        raise ValueError(msg)
    if tsunit:
        specunit = tsunit ** 2 / baseunit
    else:
        specunit = baseunit ** -1
    return specunit



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWDatafind I/O integration for `gwpy.timeseries`."""

from __future__ import annotations

import logging
from itertools import product
from typing import TYPE_CHECKING

from gwdatafind.utils import get_default_host

from ...detector import (
    Channel,
    ChannelList,
)
from ...io import datafind as io_datafind
from ...time import to_gps
from .. import (
    StateVector,
    StateVectorDict,
    TimeSeries,
    TimeSeriesDict,
)
from .losc import _any_gwosc_channels

if TYPE_CHECKING:
    from collections.abc import (
        Collection,
        Iterable,
    )
    from re import Pattern

    from ...time import SupportsToGps
    from .. import (
        TimeSeriesBase,
        TimeSeriesBaseDict,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

try:
    GWDATAFIND_SERVER = get_default_host()
except ValueError:
    GWDATAFIND_SERVER = None

DEFAULT_GWDATAFIND_SERVERS = {
    "local": GWDATAFIND_SERVER,
    "igwn": "datafind.igwn.org",
    "gwosc": "datafind.gwosc.org",
}

logger = logging.getLogger(__name__)


def find_series(
    channel: str | Channel,
    start: SupportsToGps,
    end: SupportsToGps,
    *,
    observatory: str | None = None,
    frametype: str | None = None,
    frametype_match: str | Pattern | None = None,
    host: str | None = None,
    urltype: str | None = "file",
    ext: str = "gwf",
    pad: float | None = None,
    scaled: bool | None = None,
    allow_tape: bool | None = None,
    parallel: int = 1,
    verbose: bool | str = False,
    series_class: type[TimeSeriesBase] = TimeSeries,
    **readargs,
) -> TimeSeriesBase:
    """Find and return data for a single channel using GWDataFind.

    This method uses :mod:`gwdatafind` to discover the URLs
    that provide the requested data, then reads those files using
    :meth:`TimeSeries.read()`.

    Parameters
    ----------
    channel : `str`, `Channel`
        Name of data channel to find.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS start time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine

    observatory : `str`, optional
        The observatory to use when searching for data.
        Default is to use the observatory from the channel name prefix,
        but this should be specified when searching for data in a
        multi-observatory dataset (e.g. `observatory='HLV'`).

    frametype : `str`, optional
        Name of frametype (dataset) in which this channel is stored.
        Default is to search all available datasets for a match, which
        can be very slow.

    frametype_match : `str`, optional
        Regular expression to use for frametype matching.

    host : `str`, optional
        Default is set by `gwdatafind.utils.get_default_host`.

    urltype : `str`, optional
        The URL type to use.
        Default is "file" to use paths available on the file system.

    ext : `str`, optional
        The file extension for which to search.
        "gwf" is the only file extension supported, but this may be
        extended in the future.

    pad : `float`, optional
        Value with which to fill gaps in the source data,
        by default gaps will result in a `ValueError`.

    scaled : `bool`, optional
        Apply slope and bias calibration to ADC data, for non-ADC data
        this option has no effect.

    parallel : `int`, optional
        Number of parallel threads to use when reading data.

    allow_tape : `bool`, optional
        Allow reading from frame files on (slow) magnetic tape.

    series_class : `type`, optional
        The type to use for each `Series` instance.
        Default is `TimeSeries`.

    verbose: `bool`, optional
        Use debug-level logging for data access progress.
        If ``verbose`` is specified as a string, this also defines
        the prefix for the progress meter when reading data.

    readargs
        Any other keyword arguments to be passed to `.read()`.

    Raises
    ------
    requests.exceptions.HTTPError
        If the GWDataFind query fails for any reason.

    RuntimeError
        If no files are found to read, or if the read operation
        fails.
    """
    return find_dict(
        [channel],
        start,
        end,
        observatory=observatory,
        frametype=frametype,
        frametype_match=frametype_match,
        host=host,
        urltype=urltype,
        ext=ext,
        pad=pad,
        scaled=scaled,
        allow_tape=allow_tape,
        parallel=parallel,
        verbose=verbose,
        series_class=series_class,
        **readargs,
    )[channel]


def find_dict(
    channels: Collection[str | Channel],
    start: SupportsToGps,
    end: SupportsToGps,
    *,
    observatory: str | None = None,
    frametype: str | dict[str | Channel, str | None] | None = None,
    frametype_match: str | Pattern | None = None,
    host: str | None = None,
    urltype: str | None = "file",
    ext: str = "gwf",
    pad: float | None = None,
    scaled: bool | None = None,
    allow_tape: bool | None = None,
    cache: bool | None = None,
    parallel: int = 1,
    verbose: bool | str = False,
    series_class: type[TimeSeriesBase] = TimeSeries,
    **readargs,
) -> TimeSeriesBaseDict:
    """Find and return data for multiple channels using GWDataFind.

    This method uses :mod:`gwdatafind` to discover the URLs
    that provide the requested data, then reads those files using
    :meth:`TimeSeriesDict.read()`.

    Parameters
    ----------
    channels : `list`
        List of names of data channels to find.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS start time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine

    observatory : `str`, optional
        The observatory to use when searching for data.
        Default is to use the observatory from the channel name prefix,
        but this should be specified when searching for data in a
        multi-observatory dataset (e.g. `observatory='HLV'`).

    frametype : `str`, `dict`, optional
        Name of frametype (dataset) in which this channel is stored.
        Default is to search all available datasets for a match, which
        can be very slow.
        A `dict` can be provided mapping each input channel to a specific
        frametype, with `None` (or missing) values indicating channels
        for which the frametype should be determined automatically.

    frametype_match : `str`, optional
        Regular expression to use for frametype matching.

    host : `str`, optional
        Name of the GWDataFind server to use.
        Default is set by `gwdatafind.utils.get_default_host`.

    urltype : `str`, optional
        The URL type to use.
        Default is "file" to use paths available on the file system.

    ext : `str`, optional
        The file extension for which to search.
        "gwf" is the only file extension supported, but this may be
        extended in the future.

    pad : `float`, optional
        Value with which to fill gaps in the source data,
        by default gaps will result in a `ValueError`.

    scaled : `bool`, optional
        Apply slope and bias calibration to ADC data, for non-ADC data
        this option has no effect.

    parallel : `int`, optional
        Number of parallel threads to use when reading data.

    allow_tape : `bool`, optional
        Allow reading from frame files on (slow) magnetic tape.

    cache : `bool`, `None`, optional
        Whether to cache the contents of remote URLs.
        Default (`None`) is to check the ``GWPY_CACHE`` environment variable.
        See :ref:`gwpy-env-variables` for details.

    series_class : `type`, optional
        The type to use for each `Series` instance.
        Default is `TimeSeries`.

    verbose: `bool`, optional
        Use debug-level logging for data access progress.
        If ``verbose`` is specified as a string, this also defines
        the prefix for the progress meter when reading data.

    readargs
        Any other keyword arguments to be passed to `.read()`.

    Raises
    ------
    requests.exceptions.HTTPError
        If the GWDataFind query fails for any reason.

    RuntimeError
        If no files are found to read, or if the read operation
        fails.
    """
    start = to_gps(start)
    end = to_gps(end)

    dict_class = series_class.DictClass

    # -- find frametype(s)

    groups: dict[str, list[str]] = {}
    if isinstance(frametype, dict):
        frametypes = frametype
    else:
        frametypes = dict.fromkeys(channels, frametype)
    missing = [c for c in channels if not frametypes.get(c)]

    if missing:
        logger.debug("Finding frametypes for %d channels", len(missing))
        frametypes |= io_datafind.find_best_frametype(
            missing,
            start,
            end,
            host=host,
            ext=ext,
            urltype=urltype,
            frametype_match=frametype_match,
            allow_tape=bool(allow_tape),
            cache=cache,
        )

    # flip dict to frametypes with a list of channels
    for name, ftype in frametypes.items():
        if ftype is None:
            msg = f"Cannot determine frametype for channel '{name}'"
            raise RuntimeError(msg)
        groups.setdefault(ftype, []).append(name)  # type: ignore[arg-type]

    logger.debug("Determined %s frametypes to read", len(groups))

    # -- read data

    out = dict_class()
    for ftype, clist in groups.items():
        logger.debug(
            "Reading %d channels from '%s': %s",
            len(clist),
            ftype,
            ", ".join(clist),
        )

        # parse as a ChannelList
        channellist = ChannelList.from_names(*clist)
        # strip trend tags from channel names
        names = [c.name for c in channellist]

        # find observatory for this group
        if observatory is None:
            try:
                obs = "".join(
                    sorted({c.ifo[0] for c in channellist}),
                )
            except TypeError as exc:
                msg = "Cannot parse list of IFOs from channel names"
                raise ValueError(msg) from exc

        else:
            obs = observatory

        # find frames
        urls = io_datafind.find_urls(
            obs,
            ftype,
            start,
            end,
            host=host,
            ext=ext,
            urltype=urltype,
            on_gaps="error" if pad is None else "warn",
        )
        if not urls:
            msg = (
                f"No {observatory}-{ftype} URLs found for "
                f"[{start}, {end})"
            )
            raise RuntimeError(msg)

        # read data
        new = dict_class.read(
            urls,
            names,
            start=start,
            end=end,
            pad=pad,
            scaled=scaled,
            parallel=parallel,
            verbose=verbose,
            cache=cache,
            **readargs,
        )

        # map back to user-given channel name and append
        out.append(type(new)(
            (key, new[chan]) for (key, chan) in zip(clist, names, strict=True)
        ))

    return out


# -- get registry --------------------

def _as_tuple(
    value: str | Iterable[str] | None,
    default: Iterable[str],
) -> tuple[str, ...]:
    if value is None:
        return tuple(default)
    if isinstance(value, str):
        return (value,)
    return tuple(value)


def identify_gwdatafind_sources(
    origin: str,
    channels: str | Channel | Iterable[str | Channel],
    *args,  # noqa: ARG001
    host: str | None = None,
    urltype: str | None = None,
    ext: str | None = None,
    **kwargs,  # noqa: ARG001
) -> Iterable[dict[str, object]] | None:
    """Identify GWDataFind sources and options based on the input arguments.

    This function is registered with the `get` registry for the relevant
    timeseries classes, and is called to determine what gwdatafind endpoints
    and options can be used to get the requested data.
    """
    # If not called for "get", then not applicable
    if origin != "get":
        return None

    # If the host is an NDS server, don't get involved
    if str(host).startswith("nds"):
        return []

    # Make channels a list
    channels = (channels,) if isinstance(channels, str | Channel) else channels

    # Just IFO names, not channels, most likely a GWOSC request
    if all(len(str(c)) <= 2 for c in channels):
        return []

    # Now iterate over possible gwdatafind sources by combining known options
    # for host, urltype, and extension:

    sources: list[dict[str, object]] = []

    gwosc = _any_gwosc_channels(channels)
    hosts = _as_tuple(host, filter(None, DEFAULT_GWDATAFIND_SERVERS.values()))
    urltypes = _as_tuple(urltype, ("file", "osdf"))
    exts = _as_tuple(ext, ("gwf",))
    for host_, urltype_, ext_ in product(
        hosts,
        urltypes,
        exts,
    ):
        if gwosc and host_ == DEFAULT_GWDATAFIND_SERVERS["gwosc"]:
            # Prioritise the GWOSC server for GWOSC channels
            priority = 5
        else:
            priority = 10
        sources.append({
            "host": host_,
            "urltype": urltype_,
            "ext": ext_,
            "priority": priority,
        })

    return sources


for klass, find in (
    (TimeSeries, find_series),
    (TimeSeriesDict, find_dict),
    (StateVector, find_series),
    (StateVectorDict, find_dict),
):
    klass.get.registry.register_identifier(
        "gwdatafind",
        klass,
        identify_gwdatafind_sources,
    )
    klass.get.registry.register_reader(
        "gwdatafind",
        klass,
        find,
    )



# Copyright (c) 2016-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.timeseries

Filtering a `TimeSeries` to detect gravitational waves
######################################################

The raw 'strain' output of the LIGO detectors is recorded as a `TimeSeries`
with contributions from a large number of known and unknown noise sources,
as well as possible gravitational wave signals.

In order to uncover a real signal we need to filter out noises that otherwise
hide the signal in the data. We can do this by using the :mod:`gwpy.signal`
module to design a digital filter to cut out low and high frequency noise, as
well as notch out fixed frequencies polluted by known artefacts.
"""

# %%
# Data access
# -----------
# First we download the raw strain data from the GWOSC public archive:

from gwpy.timeseries import TimeSeries
hdata = TimeSeries.get("H1", 1126259446, 1126259478)

# %%
# Data conditioning
# -----------------
# We can design a zero-pole-gain filter to remove the extranious noise.
#
# First we import the `gwpy.signal.filter_design` module and create a
# :meth:`~gwpy.signal.filter_design.bandpass` filter to remove both low and
# high frequency content

from gwpy.signal import filter_design
bp = filter_design.bandpass(50, 250, hdata.sample_rate)

# %%
# Now we want to combine the bandpass with a series of
# :meth:`~gwpy.signal.filter_design.notch` filters, so we create those
# for the first three harmonics of the 60 Hz AC mains power:

notch_frequencies = [
    60,
    120,
    180,
]
notches = [filter_design.notch(f, hdata.sample_rate) for f in notch_frequencies]

# %%
# and concatenate each of our filters together to create a single ZPK:

zpk = filter_design.concatenate_zpks(bp, *notches)

# %%
# Now, we can apply our combined filter to the data, using `filtfilt=True`
# to filter both backwards and forwards to preserve the correct phase
# at all frequencies

hfilt = hdata.filter(zpk, filtfilt=True)

# %%
# We can quickly visualise the filtered data using the
# :meth:`~TimeSeries.plot` method:

plot = hfilt.plot(figsize=[12, 4])
plot.show()

# %%
# The :mod:`~gwpy.signal.filter_design` methods return infinite impulse
# response filters by default, which, when applied, corrupt a small amount of
# data at the beginning and the end of our original `TimeSeries`.
# We can discard those data using the :meth:`~TimeSeries.crop` method
# (for consistency we apply this to both data series):

hdata = hdata.crop(*hdata.span.contract(1))
hfilt = hfilt.crop(*hfilt.span.contract(1))

# %%
# Visualisation
# -------------
# Finally, we can :meth:`~TimeSeries.plot` the original and filtered data,
# adding some code to prettify the figure:

from gwpy.plot import Plot
plot = Plot(
    hdata,
    hfilt,
    figsize=[12, 6],
    separate=True,
    sharex=True,
    color="gwpy:ligo-hanford",
)
ax1, ax2 = plot.axes
ax1.set_title("LIGO-Hanford strain data around GW150914")
ax1.text(
    1.0,
    1.01,
    "Unfiltered data",
    transform=ax1.transAxes,
    ha="right",
)
ax1.set_ylabel("Amplitude [strain]", y=-0.2)
ax2.set_ylabel("")
ax2.text(
    1.0,
    1.01,
    r"50-250\,Hz bandpass, notches at 60, 120, 180 Hz",
    transform=ax2.transAxes,
    ha="right",
)
plot.show()

# %%
# We see now a spike around 16 seconds into the data, so let's zoom into
# that time (and prettify):

plot = hfilt.plot(color="gwpy:ligo-hanford")
ax = plot.gca()
ax.set_title("LIGO-Hanford strain data around GW150914")
ax.set_ylabel("Amplitude [strain]")
ax.set_xlim(1126259462, 1126259462.6)
ax.set_xscale("seconds", epoch=1126259462)
plot.show()

# %%
# Congratulations, you have succesfully filtered LIGO data to uncover the
# first ever directly-detected gravitational wave signal, GW150914!
#
# Adding the second LIGO detector
# -------------------------------
# But wait, what about LIGO-Livingston?
# We can easily add that to our figure by following the same procedure.
#
# First, we load the new data

ldata = TimeSeries.fetch_open_data("L1", 1126259446, 1126259478)
lfilt = ldata.filter(zpk, filtfilt=True)

# %%
# The article announcing the detector told us that the signals were
# separated by 6.9 ms between detectors, and that the relative orientation
# of those detectors means that we need to invert the data from one
# before comparing them, so we apply those corrections:

lfilt.shift("6.9ms")
lfilt *= -1

# %%
# and finally make a new plot with both detectors:

plot = Plot(figsize=[12, 4])
ax = plot.gca()
ax.plot(hfilt, label="LIGO-Hanford", color="gwpy:ligo-hanford")
ax.plot(lfilt, label="LIGO-Livingston", color="gwpy:ligo-livingston")
ax.set_title("LIGO strain data around GW150914")
ax.set_xlim(1126259462, 1126259462.6)
ax.set_xscale("seconds", epoch=1126259462)
ax.set_ylabel("Amplitude [strain]")
ax.set_ylim(-1e-21, 1e-21)
ax.legend()
plot.show()

# %%
# The above filtering is (almost) the same as what was applied to LIGO data
# to produce Figure 1 in
# `Abbott et al. (2016) <https://doi.org/10.1103/PhysRevLett.116.061102>`_
# (the joint LSC-Virgo publication announcing this detection).



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.hacr`."""

from unittest import mock

import numpy
import pytest
from astropy.table import Table

from ...testing.utils import assert_table_equal
from ..filter import filter_table
from ..io import hacr as io_hacr
from ..table import EventTable
from .utils import random_table

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

requires_db_libraries = pytest.mark.requires(
    "sqlalchemy",
    "pandas",
    "pymysql",
)


@pytest.fixture(scope="module")
def hacr_table():
    """Create a table of HACR-like data."""
    names, types = zip(*{  # correct as of Jan 2025 (DMM)
        "refId": "int64",
        "freq_central": "float64",
        "bandwidth": "float64",
        "duration": "float64",
        "num_pixels": "int64",
        "snr": "float64",
        "pksnr": "float64",
        "cluster_type": "U4",
        "cluster_density": "float64",
        "segmentNumber": "int64",
        "maxPower": "float64",
        "totPower": "float64",
        "veto": "int64",
    }.items(), strict=False)
    table = random_table(
        length=100,
        names=names,
        dtypes=list(types),
    )

    # add timing information
    table.add_column(
        numpy.linspace(0, 10, num=len(table), dtype="int64", endpoint=False),
        name="gps_start",
    )
    table.add_column(
        numpy.tile(
            numpy.linspace(0, 1, num=10, dtype="float64", endpoint=False),
            10,
        ),
        name="gps_offset",
    )

    # add process_id with ints
    table.add_column(
        numpy.linspace(0, 10, num=len(table), dtype="int64", endpoint=False),
        name="process_id",
    )

    return table


@pytest.fixture(scope="module")
def hacr_process_table():
    """Create a HACR process table."""
    return Table(
        rows=[
            (0, "X1:HACR-1", 0, 1, "chacr"),
            (1, "X1:HACR-1", 1, 2, "chacr"),
            (2, "X1:HACR-1", 2, 3, "chacr"),
            (3, "X1:HACR-1", 3, 4, "chacr"),
            (4, "X1:HACR-1", 4, 5, "chacr"),
            (5, "X1:HACR-1", 5, 6, "chacr"),
            (6, "X1:HACR-1", 6, 7, "chacr"),
            (7, "X1:HACR-1", 7, 8, "chacr"),
            (8, "X1:HACR-1", 8, 9, "chacr"),
            (9, "X1:HACR-2", 9, 10, "chacr"),
            (10, "X1:HACR-3", 9, 10, "something else"),
            (11, "X1:HACR-3", 9, 10, "something else"),
        ],
        names=(
            "process_id",
            "channel",
            "gps_start",
            "gps_stop",
            "monitorName",
        ),
    )


@pytest.fixture(scope="module")
def hacr_sqlite(hacr_table, hacr_process_table, tmp_path_factory):
    """Create a HACR database in SQLite."""
    from sqlalchemy import create_engine

    db = tmp_path_factory.mktemp("hacr") / "hacr.sqlite"
    dburl = f"sqlite:///{db}"

    engine = create_engine(dburl)
    hacr_table.to_pandas().to_sql("mhacr", engine)
    hacr_process_table.to_pandas().to_sql("job", engine)
    engine.dispose()
    return dburl


@pytest.fixture(scope="module")
def hacr_engine(hacr_sqlite):
    """Create an `sqlalchemy.Engine` connected to the HACR SQLite database."""
    from sqlalchemy import create_engine

    return create_engine(hacr_sqlite)


@pytest.mark.parametrize(("start", "end", "result"), [
    ("Jan 1 2024", "Jan 1 2024 00:01", ["geo202401"]),
    ("Jan 1 2024", "Mar 1 2024", ["geo202401", "geo202402"]),
])
def test_get_database_names(start, end, result):
    """Test `get_database_names`."""
    assert io_hacr.get_database_names(start, end) == result


@requires_db_libraries
def test_get_hacr_channels(hacr_engine):
    """Test `get_hacr_channels`."""
    with mock.patch(
        "gwpy.table.io.hacr.create_engine",
        return_value=hacr_engine,
    ):
        assert set(io_hacr.get_hacr_channels()) == {
            "X1:HACR-1",
            "X1:HACR-2",
        }


@requires_db_libraries
def test_fetch_hacr(hacr_table, hacr_engine):
    """Test `EventTable.fetch(source='hacr')` with basic options."""
    t2 = EventTable.fetch(
        source="hacr",
        engine=hacr_engine,
        index_col="index",  # test sqlite database injects an index
    )

    # check response is correct
    assert_table_equal(t2, hacr_table)


@requires_db_libraries
def test_fetch_hacr_columns_where(hacr_table, hacr_engine):
    """Test that `EventTable.fetch(source='hacr')` works with complex queries."""
    with mock.patch(
        "gwpy.table.io.hacr.create_engine",
        return_value=hacr_engine,
    ):
        t2 = EventTable.fetch(
            source="hacr",
            channel="X1:HACR-1",
            start=4,
            end=6,
            columns=[
                "bandwidth",
                "snr",
            ],
            where="freq_central>500",
        )

    expect = filter_table(
        hacr_table,
        "freq_central>500",
        "gps_start >= 4",
        "gps_start < 6",
    )["bandwidth", "snr"]

    assert_table_equal(t2, expect)



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Render command-line examples for `sphinx-gallery`.

The command-line examples are specified in a INI-format configuration file.
Each section has the following options:

"title"
    The name of the example

"description"
    The summary of what is being shown

"command"
    The argv options to pass to the function

"entry_point"
    The entry-point to call with ``command``.
    Default is `gwpy-plot`.

See `/docs/cli/examples.ini` in the GWpy project for an example.
"""

from __future__ import annotations

import importlib.metadata
import logging
import shlex
from configparser import ConfigParser
from string import Template
from textwrap import indent
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from logging import Logger
    from pathlib import Path

    from sphinx.util.logging import SphinxLoggerAdapter

logger = logging.getLogger(__name__)

CLI_TEMPLATE = Template("""
\"\"\"
${titleunderline}
${title}
${titleunderline}

${description}

.. code-block:: ${language}
    ${caption}
    ${command}
\"\"\"

# %%
# The same command can be executed directly in Python as:

${pyimport}
${pyfunction}([
    ${args},
])
""".strip())


def _render_entry_point_example(
    config: ConfigParser,
    section: str,
    outdir: Path,
    logger: Logger | SphinxLoggerAdapter = logger,
    entry_point: str = "gwpy-plot",
    filename_prefix: str = "cli_",
) -> Path:
    """Render an entry point example as RST to be processed by Sphinx."""
    # read config values (allow for multi-line definition)
    argv = config.get(section, "argv").strip().replace("\n", " ")
    title = config.get(
        section,
        "title",
        fallback=" ".join(map(str.title, section.split("-"))),
    )
    desc = config.get(section, "description", fallback="")
    caption = config.get(section, "caption", fallback="")
    if caption:  # sphinx doesn't support an empty caption
        caption = f":caption: {caption}\n"
    language = config.get(section, "language", fallback="shell")
    ep_name = config.get(section, "entry_point", fallback=entry_point)

    # format the python function
    try:
        entrypoint, = importlib.metadata.entry_points(name=ep_name)
    except ValueError as exc:
        msg = "ambiguous entry point"
        raise ValueError(msg) from exc
    pymod, pyfunc = entrypoint.value.rsplit(":", 1)
    pyimport = f"from {pymod} import {pyfunc}"

    # split argv onto multiple lines
    indentstr = " " * 4
    argstr = argv.replace(" -", f" \\\\\n{indentstr}-")
    # build command-line string for display
    cmdstr = indent(f"{entrypoint.name} {argstr}", indentstr).strip()

    # build code to generate the plot when sphinx runs
    pyargs = indent(
        # join the arguments into a list, but drop a
        # new line for each new option flag, and indent
        ", ".join(map(repr, shlex.split(argv))).replace(
            ", '-",
            ",\n'-",
        ),
        indentstr,
    ).strip()

    # create Python script
    pyfile = outdir / f"{filename_prefix}{section}.py"
    code = CLI_TEMPLATE.substitute(
        title=title,
        titleunderline="#" * len(title),
        description=desc,
        language=language,
        caption=caption,
        command=cmdstr,
        args=pyargs,
        pyimport=pyimport,
        pyfunction=pyfunc,
    )
    pyfile.write_text(code)
    logger.info("[cli] wrote %s", str(pyfile))
    return pyfile


def render_entry_point_examples(
    inifile: Path,
    outdir: Path,
    logger: Logger | SphinxLoggerAdapter = logger,
    entry_point: str = "gwpy-plot",
    filename_prefix: str = "cli_",
) -> None:
    """Render entry-point examples as RST to be processed by Sphinx."""
    outdir.mkdir(exist_ok=True, parents=True)

    # read example config
    config = ConfigParser()
    config.read(inifile)

    logger.info("rendering entry point examples")

    # render examples
    for sect in config.sections():
        _render_entry_point_example(
            config,
            sect,
            outdir,
            logger,
            entry_point=entry_point,
            filename_prefix=filename_prefix,
        )

    # write a gallery header (blank)
    (outdir / "GALLERY_HEADER.rst").write_text("""
Click on a thumbnail image below to see the command used to generate
the example, and the full-resolution output image.
""".strip())



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.spectrogram.spectrogram`."""

from __future__ import annotations

from io import BytesIO
from typing import (
    TYPE_CHECKING,
    TypeVar,
)

import numpy
import pytest
from astropy import units
from matplotlib import rc_context
from scipy import signal

from ...testing import utils
from ...types.tests.test_array2d import TestArray2D as _TestArray2D
from .. import Spectrogram

if TYPE_CHECKING:
    from ...signal.filter_design import ZpkType

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

SpectrogramType = TypeVar("SpectrogramType", bound=Spectrogram)


class TestSpectrogram(_TestArray2D[Spectrogram]):
    """Tests of `gwpy.spectrogram.Spectrogram`."""

    TEST_CLASS = Spectrogram

    def test_new(self):
        """Test creating new `Spectrogram` objects."""
        super().test_new()

        # check handling of epoch vs t0
        a = self.create(epoch=10)
        b = self.create(t0=10)
        utils.assert_quantity_sub_equal(a, b)

    def test_new_redundant_args(self):
        """Test creating new `Spectrogram` objects with redundant args."""
        with pytest.raises(
            ValueError,
            match=r"^give only one of epoch or t0$",
        ):
            self.TEST_CLASS(self.data, epoch=1, t0=1)

    def test_new_times(self):
        """Test creating new `Spectrogram` objects with ``times`` argument."""
        times = numpy.arange(self.data.shape[0])
        a = self.create(times=times)
        utils.assert_quantity_equal(a.times, times * units.second)

    def test_epoch(self, array):
        """Test `Spectrogram.epoch` property."""
        assert array.epoch.gps == array.x0.value

    def test_value_at(self):
        """Test `Spectrogram.value_at` method."""
        super().test_value_at()
        array = self.create()
        assert array.value_at(
            5000 * units.millisecond,
            2000 * units.milliHertz,
        ) == self.data[5][2] * array.unit

    @pytest.mark.parametrize("ratio", ["mean", "median"])
    def test_ratio(self, array: SpectrogramType, ratio):
        """Test `Spectrogram.ratio` method."""
        rat = array.ratio(ratio)
        array_meth = getattr(array, ratio)
        utils.assert_quantity_sub_equal(rat, array / array_meth(axis=0))

    def test_ratio_invalid_operand(self, array: SpectrogramType):
        """Test `Spectrogram.ratio` method with invalid input."""
        with pytest.raises(
            ValueError,
            match="operand 'blah' unrecognised",
        ):
            array.ratio("blah")

    @pytest.fixture
    def min_max_mean(self, array: SpectrogramType):
        """Return min, max, mean `Spectrogram` objects for testing."""
        min_ = array.min(axis=0)
        max_ = array.max(axis=0)
        mean = array.mean(axis=0)
        return min_, max_, mean

    def test_from_spectra(self, min_max_mean):
        """Test `Spectrogram.from_spectra` class method."""
        min_, max_, mean = min_max_mean

        # check basic stack works
        new = self.TEST_CLASS.from_spectra(mean, min_, max_, dt=1)
        assert new.shape == (3, min_.size)
        assert new.name == mean.name
        assert new.epoch == mean.epoch
        assert new.f0 == mean.f0
        assert new.df == mean.df
        assert new.unit == mean.unit
        assert new.dt == 1 * units.second
        utils.assert_array_equal(
            new.value,
            numpy.vstack((mean.value, min_.value, max_.value)),
        )

        # check kwargs
        new = self.TEST_ARRAY.from_spectra(
            mean,
            min_,
            max_,
            dt=2,
            epoch=0,
            f0=100,
            df=.5,
            unit="meter",
            name="test",
        )
        assert new.name == "test"
        assert new.epoch.gps == 0
        assert new.f0 == 100 * units.Hertz
        assert new.df == 0.5 * units.Hertz
        assert new.unit == units.meter

    def test_from_spectra_error_dt(self):
        """Test `Spectrogram.from_spectra` dt error handling."""
        fs = self.TEST_CLASS._columnclass(self.data[0], epoch=0, f0=0, df=1)

        # Check that dt is required with a single input
        with pytest.raises(
            ValueError,
            match="cannot determine dt",
        ):
            self.TEST_ARRAY.from_spectra(fs)
        self.TEST_ARRAY.from_spectra(fs, dt=1)

    def test_from_spectra_error_shape(self):
        """Test `Spectrogram.from_spectra` shape error handling."""
        fs = self.TEST_CLASS._columnclass(self.data[0], epoch=0, f0=0, df=1)

        with pytest.raises(
            ValueError,
            match="all the input array dimensions",
        ):
            self.TEST_CLASS.from_spectra(fs, fs[1:])
        with pytest.raises(
            ValueError,
            match="all the input array dimensions",
        ):
            self.TEST_CLASS.from_spectra(fs, fs[::2])

    def test_crop_frequencies(self, array):
        """Test `Spectrogram.crop_frequencies` method."""
        # test simple
        array2 = array.crop_frequencies()
        utils.assert_quantity_sub_equal(array, array2)
        assert numpy.may_share_memory(array.value, array2.value)

    def test_crop_frequencies_normal(self, array):
        """Test `Spectrogram.crop_frequencies` normal cropping."""
        array2 = array.crop_frequencies(2, 5)
        utils.assert_array_equal(array2.value, array.value[:, 2:5])
        assert array2.f0 == 2 * units.Hertz
        assert array2.df == array.df

    def test_crop_frequencies_copy(self, array: SpectrogramType):
        """Test `Spectrogram.crop_frequencies` copy option."""
        # test copy
        array2 = array.crop_frequencies(copy=True)
        assert not numpy.may_share_memory(array.value, array2.value)

    def crop_frequencies_warnings(self, array):
        """Test `Spectrogram.crop_frequencies` warnings."""
        # test warnings
        with pytest.warns(
            UserWarning,
            match="FIXME",
        ):
            array.crop_frequencies(array.yspan[0]-1, array.yspan[1])
        with pytest.warns(
            UserWarning,
            match="FIXME",
        ):
            array.crop_frequencies(array.yspan[0], array.yspan[1]+1)

    @pytest.mark.parametrize("method", ["imshow", "pcolormesh"])
    def test_plot(self, array, method):
        """Test `Spectrogram.plot` method."""
        with rc_context(rc={"text.usetex": False}):
            plot = array.plot(method=method)
            ax = plot.gca()
            assert len(ax.lines) == 0
            if method == "imshow":
                assert len(ax.images) == 1
            else:
                assert len(ax.collections) == 1
            assert ax.get_epoch() == array.x0.value
            assert ax.get_xlim() == array.xspan
            assert ax.get_ylim() == array.yspan
            plot.save(BytesIO(), format="png")
            plot.close()

    @pytest.fixture
    def zpk(self) -> ZpkType:
        """Return a sample zero-pole-gain filter."""
        z = numpy.asarray([0j, -1j, 1j])
        p = numpy.asarray([-0.5 + 0.5j, -0.5 - 0.5j, -2])
        k = 1
        return z, p, k

    def test_zpk(self, array, zpk):
        """Test `Spectrogram.zpk` method."""
        utils.assert_quantity_sub_equal(
            array.zpk(*zpk),
            array.filter(zpk, analog=False, unit="rad/s"),
        )

    def test_filter(self, array, zpk):
        """Test `Spectrogram.filter` method."""
        # build filter - convert to rad/s since freqresp expects rad/s
        # and filter() defaults to rad/s
        lti = signal.lti(*zpk)
        # freqresp expects omega (rad/s), convert Hz to rad/s
        omega = array.frequencies.value * 2 * numpy.pi
        fresp = numpy.nan_to_num(abs(lti.freqresp(w=omega)[1]))

        # test simple filter
        a2 = array.filter(zpk, analog=True)
        utils.assert_quantity_sub_equal(array * fresp, a2)

    def test_filter_inplace(self, array, zpk):
        """Test `Spectrogram.filter(..., inplace=True)`."""
        a2 = array.filter(zpk, inplace=False, analog=True)
        array.filter(zpk, inplace=True, analog=True)
        utils.assert_quantity_sub_equal(array, a2)

    def test_filter_bad_kwargs(self, array, zpk):
        """Test `Spectrogram.filter` method."""
        # test errors
        with pytest.raises(
            TypeError,
            match=r"got an unexpected keyword argument 'blah'",
        ):
            array.filter(zpk, blah=1)

    def test_read_write_hdf5(self):
        """Test HDF5 read/write of `Spectrogram` objects."""
        array = self.create(name="X1:TEST")
        utils.test_read_write(
            array,
            "hdf5",
            write_kw={"overwrite": True},
        )

    def test_percentile(self):
        """Test `Spectrogram.percentile` method."""
        array = self.create(name="Test", unit="m")
        a2 = array.percentile(50)
        utils.assert_quantity_sub_equal(
            array.median(axis=0),
            a2,
            exclude=("name",),
        )
        assert a2.name == "Test: 50th percentile"
        assert a2.unit == array.unit
